import unittest
from tios import gromacs_utilities
import tempfile
import os
import zlib
import mdtraj as mdt

class TestGromacsUtilitiesMethods(unittest.TestCase):

    def test_complete_inputs_nothing_needed(self):
        inputs = {'-s': 'test.tpr'}
        new_inputs = gromacs_utilities.complete_inputs(inputs)
        self.assertDictEqual(inputs, new_inputs)

    def test_complete_inputs_with_deffnm(self):
        inputs = {'-deffnm': 'test'}
        expected_inputs = { '-s'      : 'test.tpr',
                            '-o'      : 'test.trr',
                            '-x'      : 'test.xtc',
                            '-cpo'    : 'test.cpt',
                            '-e'      : 'test.edr',
                            '-g'      : 'test.log'}
        new_inputs = gromacs_utilities.complete_inputs(inputs)
        self.assertDictEqual(expected_inputs, new_inputs)

    def test_complete_inputs_for_imd(self):
        inputs = {'-deffnm': 'test'}
        expected_inputs = { '-deffnm' : 'test',
                            '-imdwait': None,
                            '-imdterm': None,
                            '-imdport': '40237'}
        new_inputs = gromacs_utilities.complete_inputs_for_imd(inputs)
        self.assertDictEqual(expected_inputs, new_inputs)

    def test_apply_checkpoint(self):
        testdir = tempfile.mkdtemp()
        checkpoint = zlib.compress(b'Dummy checkpoint data')
        inputs = {'-deffnm': '{}/test'.format(testdir)}
        expected_inputs = {'-deffnm' : '{}/test'.format(testdir),
                           '-cpi' : '{}/test.cpt'.format(testdir),
                           '-noappend' : None}
        new_inputs = gromacs_utilities.apply_checkpoint(inputs, checkpoint)
        self.assertEqual(new_inputs, expected_inputs)
        self.assertTrue(os.path.exists('{}/test.cpt'.format(testdir)))
        
    def test_tpr_file_valid_for_imd(self):
        tprfile = 'test/examples/bpti.tpr'
        self.assertTrue(gromacs_utilities.tpr_file_valid_for_imd(tprfile))

    def test_tpr_file_does_not_set_imd(self):
        tprfile = 'test/examples/bpti_no_imd.tpr'
        self.assertFalse(gromacs_utilities.tpr_file_valid_for_imd(tprfile))

    def test_tpr_file_is_NPT(self):
        tprfile = 'test/examples/bpti_NPT.tpr'
        self.assertFalse(gromacs_utilities.tpr_file_valid_for_imd(tprfile))

    def test_dt_from_tpr_file(self):
        tprfile = 'test/examples/bpti.tpr'
        dt = gromacs_utilities.dt_from_tpr_file(tprfile)
        self.assertEqual(dt, 0.002)

    def test_trajectory_from_tpr_file(self):
        tprfile = 'test/examples/bpti.tpr'
        traj = gromacs_utilities.trajectory_from_tpr_file(tprfile)
        self.assertIsInstance(traj, mdt.Trajectory)
        self.assertEqual(traj.topology.n_atoms, 20521)

    def test_step_from_checkpoint_file(self):
        cptfile = 'test/examples/bpti.cpt'
        step = gromacs_utilities.step_from_checkpoint_file(cptfile)
        self.assertEqual(step, 520)



import unittest
from tios import utilities

class TestCheckGromacs514Methods(unittest.TestCase):

    def test_check__installed_version(self):
        result = utilities.installed_version('NAMD')
        self.assertEqual(result, '2.12')



import unittest
from tios import receiver, transmitter
import shutil
import tempfile
import os
import mdtraj as mdt

class TestTiosReceiverMethods(unittest.TestCase):

    def setUp(self):
        self.protocol = 'Mongo'
        self.testdir = 'xxxxxx'
        os.mkdir(self.testdir)
        shutil.copy('test/examples/bpti.tpr', self.testdir)
        shutil.copy('test/examples/bpti.cpt', self.testdir)

    def tearDown(self):
        shutil.rmtree(self.testdir)

    def test_initialize_new_tios_receiver_from_database(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        id = tj.id

        tj2 = receiver.TiosReceiver(id, protocol=self.protocol)
        self.assertIsInstance(tj2, receiver.TiosReceiver)
        self.assertEqual(tj2.status, 'Ready')
        self.assertIsInstance(tj2.trajectory, mdt.Trajectory)
        self.assertEqual(tj2.trajectory.n_atoms, 20521)

        tj3 = receiver.TiosReceiver(id, protocol=self.protocol, selection='protein')
        self.assertEqual(tj3.trajectory.n_atoms, 892)

    def test_tios_receiver_collection(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        id = tj.id

        tj2 = receiver.TiosReceiver(id, protocol=self.protocol)
        self.assertEqual(tj2.status, 'Ready')
        tj.start()
        self.assertEqual(tj.status, 'Running')
        self.assertEqual(len(tj2.trajectory), 1)
        tj.step()
        self.assertEqual(tj.status, 'Running')
        tj2.step()
        self.assertEqual(tj2.status, 'Running')
        self.assertEqual(len(tj2.trajectory), 1)
        tj.step()
        tj2.step()
        self.assertEqual(len(tj2.trajectory), 1)
        tj.stop()
        self.assertEqual(tj2.status, 'Stopped')
        self.assertEqual(tj.status, 'Stopped')
      



import unittest
from tios import transmitter, communication
import shutil
import tempfile
import mdtraj as mdt

from contextlib import contextmanager
import os

@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

class TestTiosNAMDTransmitterMethods(unittest.TestCase):

    def setUp(self):
        #self.protocol = 'Dummy'
        self.protocol = 'Mongo'
        #self.testdir = tempfile.mkdtemp()
        self.testdir = 'scratchdir'
        if os.path.exists(self.testdir):
            shutil.rmtree(self.testdir)
        os.mkdir(self.testdir)
        shutil.copy('test/examples/ubq_wb.conf', self.testdir)
        shutil.copy('test/examples/ubq_wb.psf', self.testdir)
        shutil.copy('test/examples/ubq_wb.pdb', self.testdir)
        shutil.copy('test/examples/ubq_wb_eq.coor', self.testdir)
        shutil.copy('test/examples/ubq_wb_eq.xsc', self.testdir)
        shutil.copy('test/examples/par_all27_prot_lipid.inp', self.testdir)

    def tearDown(self):
        shutil.rmtree(self.testdir)

    def test_initialize_new_tios_namd_transmitter_from_command(self):
        with cd(self.testdir):
            tj = transmitter.from_command('namd2 +p4 ubq_wb.conf', 
                                          title='NAMD test',
                                          selection='name CA', 
                                          protocol = self.protocol)
            self.assertIsInstance(tj, transmitter.TiosTransmitter)

    def test_initialize_new_tios_namd_transmitter_from_database(self):
        tj = transmitter.from_command('namd2 +p4 test/examples/ubq_wb.conf', 
                                      title='NAMD test',
                                      selection='name CA', 
                                      protocol=self.protocol)
        id = tj.id

        with cd(self.testdir):
            tj2 = transmitter.from_entry(id, protocol=self.protocol)
            self.assertIsInstance(tj2, transmitter.TiosTransmitter)

    def test_tios_namd_job_from_command_datastore_integrity(self):
        tj = transmitter.from_command('namd2 +p4 test/examples/ubq_wb.conf', 
                                      title='NAMD test',
                                      selection='name CA', 
                                      protocol=self.protocol)
        self.assertEqual(tj._te.title, 'NAMD test')
        self.assertEqual(tj._te.timepoint, 0.0)
        self.assertEqual(tj._te.md_code, 'NAMD')
        self.assertEqual(tj._te.status, 'Ready')
        self.assertEqual(tj._te.trate, 1.0)
        self.assertEqual(tj._te.frame_rate, 0.0)
        self.assertIsInstance(tj._te.topology, mdt.Topology)
        self.assertEqual(len(tj._te.selection), 76)
        self.assertEqual(tj._te.xyzsel.shape, (76, 3))

    def test_tios_namd_job_from_command_datastore_integrity2(self):
        tj = transmitter.from_command('namd2 +p4 test/examples/ubq_wb.conf', 
                                   title='NAMD test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        id = tj.id

        with cd(self.testdir):
            tj2 = transmitter.from_entry(id, protocol=self.protocol)

            self.assertEqual(tj2._te.title, 'NAMD test')
            self.assertEqual(tj2._te.timepoint, 0.0)
            self.assertEqual(tj2._te.md_code, 'NAMD')
            self.assertEqual(tj2._te.status, 'Ready')
            self.assertEqual(tj2._te.trate, 1.0)
            self.assertEqual(tj2._te.frame_rate, 0.0)
            self.assertIsInstance(tj2._te.topology, mdt.Topology)
            self.assertEqual(len(tj2._te.selection), 76)
            self.assertEqual(tj2._te.xyzsel.shape, (76, 3))

    def test_tios_namd_transmitter_start_step_and_stop(self):
        with cd(self.testdir):
            tj = transmitter.from_command('namd2 +p4 ubq_wb.conf', 
                                          title='NAMD test',
                                          selection='name CA', 
                                          protocol=self.protocol)
            tj.start()
            self.assertEqual(tj._te.status, 'Running')
            self.assertFalse(tj._imd_sim.new_checkpoint)
            tj.step()
            self.assertEqual(tj._te.timepoint, 1.00)
            tj.step()
            self.assertEqual(tj._te.timepoint, 2.00)
            tj.stop()
            self.assertEqual(tj._te.status, 'Stopped')
        
    def test_tios_namd_transmitter_start_stop_and_restart(self):
        with cd(self.testdir):
            tj = transmitter.from_command('namd2 +p4 ubq_wb.conf', 
                                          title='NAMD test',
                                          selection='name CA', 
                                          protocol=self.protocol)
            tj.start()
            tj.step()
            tj.step()
            tj.stop()
            self.assertAlmostEqual(tj._te.timepoint, 2.000)
            id = tj.id

        with cd('scratchdir'):
            tj2 = transmitter.from_entry(id, protocol=self.protocol)
            tj2.start()
            self.assertEqual(tj2._te.status, 'Running')
            tj2.step()
            tj2.stop()
            self.assertEqual(tj2._te.status, 'Stopped')

    def test_tios_namd_transmitter_stop_by_message(self):
        with cd(self.testdir):
            tj = transmitter.from_command('namd2 +p4 ubq_wb.conf', 
                                          title='NAMD test',
                                          selection='name CA', 
                                          protocol=self.protocol)
            tj.start()
            tj.step()
            tj.step()
            self.assertEqual(tj._te.status, 'Running')
            tj._te.message = {'header' : 'STOP'}
            tj._te.sync()
            tj.step()
            self.assertEqual(tj._te.status, 'Stopped')




'''
Tios_transmitter: runs an MD code and streams the data using the tios
protocol.

A tios transmitter can be created from a command-line style string:

    tt = from_command('gmx mdrun -deffnm test')

or from an existing entry in the tios database:

    tt = from_entry('abc123')

The simulation is launched via the start method:

    tt.start()

Being under IMD control, it is advanced to the next sampling interval
using the step method:

    tt.step()

And stopped using the stop method:

    tt.stop()

To check the simulation is still running, use the is_running method:

    alive = tt.is_running()

Once started, the standard output and standard error from the job are
available:

    print tt.stdout
    print tt.stderr

An MDtraj Trajectory object with a single frame - the current coordinates - 
is available:

    topology = tt.trajectory.topology
    crds = tt.trajectory.xyz

Values of certain simulation metrics - e.g. temperature, potential energy, 
are available via the monitors dictionary attribute.  The available keys 
are: 'Timepoint', 'T', 'Etot', 'Epot', 'Evdw', 'Eelec', 'Ebond', 'Eangle', 
'Edihe', 'Eimpr'.

'''
from tios import utilities, communication, imd_sim

import socket
import os
import numpy as np
import time
from collections import deque
import mdtraj as mdt

def from_command(command, title='', selection='all', trate=1.0, 
                 deque_length=50, protocol='Mongo'):
    """
    Create a Tios transmitter from a command line string.

    Args:
        command (str): command-line like string.
        title (str, optional): Descriptive title for the job.
        selection (str, optional): Atom selection in MDTraj syntax. These 
           atoms are stored separately in the database from the rest, 
           potentially reducing data transfer to the receiver.
        trate (float, optional): The IMD sampling interval in picoseconds.
            Defaults to 1ps.
        deque_length (int, optional): Length of data timeseries to retain.
            Defaults to 50.
        protocol (str, optional): Communication protocol to use. Defaults to
            "Mongo".

    Returns:
        An instance of a `TiosTransmitter`
    """
    imd_job = imd_sim.from_command(command)
    id = utilities.id_generator()
    is_duplicate = True
    while is_duplicate:
        try:
            te = communication.TiosAgent(id, protocol=protocol, new=True)
            is_duplicate = False
        except ValueError:
            id = utilities.id_generator()
    id_const = id + '_const'
    te_const = communication.TiosAgent(id_const, protocol=protocol, new=True)
    te.title = title
    te.trate = trate
    te.message = {'header' : None}
    te.md_code = utilities.md_code_from_string(command)
    inputs = utilities.string_to_inputs_dict(command)
    te_const.filepack = utilities.inputs_dict_to_filepack(inputs)
    te_const.sync()
    te.checkpoint = None
    topology = imd_job.trajectory.topology
    sel = topology.select(selection).tolist()
    te.topology = topology
    te.selection = sel
    unsel = [i for i in range(topology.n_atoms) if not i in sel]
    te.box = imd_job.trajectory.unitcell_vectors
    te.xyzsel = imd_job.trajectory.xyz[:,sel][0]
    te.xyzunsel = imd_job.trajectory.xyz[:,unsel][0]
    monitors = {}
    for key in ['Timepoint', 'T', 'Etot', 'Epot', 'Evdw', 'Eelec', 'Ebond', 
                'Eangle', 'Edihe', 'Eimpr', 'RMSD']:
        monitors[key] = None
    te.monitors = monitors
    te.sync()
    return TiosTransmitter(imd_job, te)

def from_entry(id, protocol='Mongo', force=False,
               targetdir='', preamble=None, extra_args=None, trate=None):
    """
    Create a Tios transmitter from an entry in the Tios database.

    Args:
        id (str): The Tios ID.
        protocol (str, optional): The connection protocol. Defaults to "Mongo".
        force (bool, optional): If True, a transmitter will be created even if
            the job appears to still be running somewhere.
        targetdir (str, optional): The directory into which files should be
            unpacked and from which the job will be launched. Defaults to the
            current directory.
        preamble (str, optional): Preamble required for the command line, e.g.
            "mpirun -n 16".
        extra_args (str, optional): Extra arguments to be appended to the
            command line, e.g. from Gromacs, "-maxh 23".
        trate (float, optional): IMD sampling interval. Defaults to whatever
            is defined in the database entry.

    Returns:
        An instance of a `TiosTransmitter`

    """
    te = communication.TiosAgent(id, protocol=protocol)
    te_const = communication.TiosAgent(id + '_const', protocol=protocol)
    if te.status == 'Running':
        if force:
            print('Warning: Tios job {} appears to be running already'.format(id))
        else:
            raise RuntimeError('Tios job {} is already running'.format(id))
    inputs = utilities.filepack_to_inputs_dict(te_const.filepack, targetdir=targetdir)
    if preamble is not None:
        inputs[-1] = preamble.split()
    if trate is None:
       trate = te.trate
    else:
       te.trate = trate
    run_command = utilities.inputs_dict_to_string(inputs)
    if extra_args is not None:
        run_command = run_command + " " + extra_args
    imd_job = imd_sim.from_command(run_command, checkpoint=te.checkpoint, trate=trate)
    te.message = {'header' : None}
    te.sync()
    return TiosTransmitter(imd_job, te)
        
class TiosTransmitter(object):
    def __init__(self, imd_sim, tios_agent):
        """
        Transmitter for an MD simulation job, as created by Tios

        Args:
            imd_sim (IMDJob): An instance of an `IMDJob`.
            tios_agent (TiosAgent): An instance of a `TiosAgent`

        Attributes:
            id (str): Tios ID for the running job.
            status (str): Status of the running job.
            trajectory (MDTraj.Trajectory): Single frame MD trajectory.
            monitors (dict): Dictionary of energy components.
            stdout (str): Contents of STDOUT from the running job.
            stderr (str): Contents of STDERR from the running job.

        """
        self._imd_sim = imd_sim
        self._te = tios_agent
        self.id = self._te.id
        self.status = imd_sim.status
        self._selection = tios_agent.selection
        self._topology = tios_agent.topology
        self._unselection = [i for i in range(self._topology.n_atoms) 
                             if not i in self._selection]
        self.monitors = self._te.monitors
        self.trajectory = imd_sim.trajectory
        self._start = mdt.Trajectory(imd_sim.trajectory.xyz, 
                                     imd_sim.trajectory.topology)
        self._te.framerate = 0.0
        self._te.md_version = utilities.installed_version(self._te.md_code)
        self._te.host = socket.gethostname()
        try:
            self._te.username = os.getlogin()
        except OSError:
            self._te.username = '(unknown)'
        self._te.frame_rate = 0.0
        self._te.timepoint = imd_sim.timepoint
        self._te.status = imd_sim.status
        self._te.sync()

    def start(self):
        """
        Launch the simulation.

        """
        self._imd_sim.start()
        self.stdout = self._imd_sim.stdout
        self.stderr = self._imd_sim.stderr
        if self._imd_sim.status != 'Running':
            print(self.stdout)
            print(self.stderr)
            raise RuntimeError('Job failed prematurely')
        self._te.status = self._imd_sim.status
        self.status = self._imd_sim.status
        self._te.sync()
        self._last_step_time = time.time()

    def stop(self):
        """
        Stop the simulation.

        """
        self._imd_sim.stop()
        self._te.status = self._imd_sim.status
        self.status = self._imd_sim.status
        if self._imd_sim.new_checkpoint:
            cpt = self._imd_sim.get_checkpoint()
            self._te.checkpoint = cpt
        self._te.sync()
        
    def step(self):
        """
        Move the simulation along to the next IMD sample point.

        """
        message = self._te.message
        if message['header'] == 'STOP':
            self.stop()
        else:
            self._imd_sim.step()
            crds = self._imd_sim.trajectory.xyz[0]
            self._te.xyzsel = crds[self._selection]
            self._te.xyzunsel = crds[self._unselection]
            energies = self._imd_sim.energies
            energies['Timepoint'] = self._imd_sim.timepoint
            if 'tstep' in energies:
                del energies['tstep']
            for key in energies:
                self.monitors[key] = energies[key]
            self._te.monitors = self.monitors
            if self._imd_sim.new_checkpoint:
                cpt = self._imd_sim.get_checkpoint()
                self._te.checkpoint = cpt
            now = time.time()
            self._te.frame_rate = 60.0 / (now - self._last_step_time)
            self._last_step_time = now
            self._te.timepoint = self._imd_sim.timepoint
            self._te.status = self._imd_sim.status
            self.status = self._imd_sim.status
            self._te.sync()

    def is_running(self):
         """
         Return True if the simulation appears to be running.

         """
         return self._imd_sim.is_running()




import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from PIL import Image, ImageDraw
import numpy as np


def plot_monitors(fname, dic, xkey, size=(8,8)):
    '''
    Make an image file from the monitors data
    '''
    x = list(dic.pop(xkey))
    xyz = dic.pop('xyz', None)
    fig, axarr = plt.subplots(len(dic), sharex=True, figsize=size)
    plt.subplots_adjust(wspace=0, hspace=0)
    xmin = float(min(x))
    xmax = float(max(x))
    axarr[0].set_xlim([xmin, xmax])
    axarr[-1].set_xlabel('time (ps)')
    for ax, key in zip(axarr, dic.keys()):
        ax.set_ylabel(key)
        y = list(dic[key])
        ymin = float(min(y))
        ymax = float(max(y))
        yrange = ymax - ymin
        ax.set_ylim([ymin - 0.1 * yrange, ymax + 0.1 * yrange])
        ax.yaxis.set_ticks([ymin, ymax])
        ax.fill_between(x, ymin - 0.1 * yrange, y)
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()

def element_colour(e):
    ec = {
        'C' : (0,0,0),
        'N' : (0, 0, 255),
        'O' : (255, 0, 0),
        'H' : (0, 255, 255),
        'S' : (255, 0, 255),
        'P' : (255, 255, 0)
        }
    if e in ec:
        return ec[e]
    else:
        return (128, 128, 128)

def plot_molecule(outfile, traj, selection=None, frame=0, size=(4,3)):
    '''
    Produce a simple wireframe image of the trajectory frame.
    '''
    imgsize = max(size) * 96 # assume 96 dpi
    if selection is not None:
        sel = traj.topology.select(selection)
        traj.topology = traj.topology.subset(sel)
        traj.xyz = traj.xyz[:, sel]
    elements = [a.element.symbol for a in traj.topology.atoms]
    bonds = [(b[0].index, b[1].index) for b in traj.topology.bonds]
    xy = traj.xyz[frame, :, :2]

    supersample = 4
    linewidth = 2

    molimg(outfile, xy, elements, bonds, imgsize=imgsize, 
           supersample=supersample, linewidth=linewidth)

def molimg(imgfile, xy, elements, bonds, imgsize=400, 
           supersample=4, linewidth=1):
    '''
    Produce a wireframe image of a molecule and save to file.
    Uses PIL rather than matplotlib as it is faster.
    '''

    xy = xy - xy.min(axis=0)
    xy = xy / xy.max()
    xy -= ((xy.max(axis=0) - 1) / 2)
    ixy = (xy * imgsize * supersample).astype(np.int)
    ims = int(imgsize * supersample)
    imgsize = int(imgsize)

    im = Image.new('RGB', (ims, ims), color=(255, 255, 255))
    draw = ImageDraw.Draw(im)
    lw = linewidth * supersample
    for b in bonds:
        start = tuple(ixy[b[0]])
        end = tuple(ixy[b[1]])
        mid = tuple((ixy[b[0]] + ixy[b[1]]) / 2)
        draw.line((start, mid), fill=element_colour(elements[b[0]]), width=lw)
        draw.line((end, mid), fill=element_colour(elements[b[1]]), width=lw)

    im = im.resize((imgsize, imgsize), resample=Image.BICUBIC)
    im.save(imgfile, 'PNG')



'''
Tios_receiver complements tios_transmitter.

A tios receiver is initialised with a tios job ID:

    tr = tios.receiver('abc123')

It collects a new frame of data every time the step() method is called:

    tr.step()

To see if the simulation is still running, use the is_running method:

    alive = tr.is_running()

NOTE: tr.step() does not return with an error if the simulation is stopped - it
just waits until it starts again.

The receiver has an MDTraj trajectory attribute with data for the current 
snapshot. This is updated each time the step() method is called.

The recent history of certain simula.tion metrics - e.g. temperature, 
potential energy, are available via the monitors dictionary attribute. Each 
key is associated with a list that grows as frames are read. The available 
keys are: 'Timepoint', 'T', 'Etot', 'Epot', 'Evdw', 'Eelec', 'Ebond', 'Eangle',
'Edihe', 'Eimpr', 'RMSD'.

    rmsd_history = tr.monitors['RMSD']

'''
from tios import communication

import mdtraj as mdt
import numpy as np
import time

class TiosReceiver(object):
    def __init__(self, id, protocol='Mongo', selection='all'):
        """
        Create a new receiver.

        Args:
            id (str): ID of the Tios job.
            protocol (str, optional): Communication protocol to use.
            selection (str, optional): selection string in MDTraj syntax for
                the subset of atoms to be included.

        Attributes:
            trajectory (MDTraj.Trajectory): Trajectory with current coordinates
            monitors (dict): Dictionary of simulation metrics (energies, etc.).
            timepoint (float): Age of the simulation, in picoseconds.

        """
        self._te = communication.TiosAgent(id, protocol=protocol)
        topology = self._te.topology
        self._box = self._te.box
        self._usersel = topology.select(selection)
        self._datasel = self._te.selection
        if set(self._usersel).issubset(set(self._datasel)):
            self._use_unsel = False
            self._sel = []
            for i in self._usersel:
                self._sel.append(self._datasel.index(i))
            self._xyz = self._te.xyzsel
        else:
            self._use_unsel = True
            self._unsel = []
            for i in range(topology.n_atoms):
                if not i in self._datasel:
                    self._unsel.append(i)
            self._xyz = np.zeros((topology.n_atoms, 3))
            self._xyz[self._datasel] = self._te.xyzsel
            self._xyz[self._unsel] = self._te.xyzunsel
            self._sel = self._usersel
        topology = topology.subset(self._usersel)
        self.trajectory = mdt.Trajectory(self._xyz[self._sel], topology)
        self.trajectory.unitcell_vectors=self._box
        self.monitors = self._te.monitors
        self.timepoint = self._te.timepoint
        self.trajectory.time = self.timepoint

    @property
    def status(self):
        """
        str: Current status of simulation.

        """
        return self._te.status

    def step(self, wait=True, killer=None):
        """
        Collect the next timestep of data from the simulation.

        Args:
            wait (bool, optional): If True, if the simulation is Stopped, wait
                for it to update. Otherwise return immediately.e
            killer (GracefulKiller, optional): hook to trap ctrl-C, etc.

        """
        newtimepoint = self.timepoint
        while newtimepoint == self.timepoint:
            if killer is not None:
                if killer.kill_now:
                    return
            if self.status != 'Running':
                if wait:
                    interval = 30
                else:
                    return
            else:
                if self._te.frame_rate == 0:
                    interval = 2.0
                else:
                    interval = max(2.0, 30.0 / self._te.frame_rate)
            time.sleep(interval)
            newtimepoint = self._te.timepoint
        self.timepoint = newtimepoint
        xyzsel = self._te.xyzsel
        if self._use_unsel:
            self._xyz[self._datasel] = xyzsel
            self._xyz[self._unsel] = self._te.xyzunsel
        else:
            self._xyz = xyzsel
        self.trajectory.xyz = self._xyz[self._sel]
        self.trajectory.time = self.timepoint
        self.monitors = self._te.monitors

    def is_running(self):
        """
        Return True if the simulation appears to be running.

        """
        return self.status == 'Running'



"""
Utilities to extract required information from NAMD files
"""
from __future__ import print_function
import sys
import os
import subprocess
import tempfile
import zlib
import mdtraj as mdt

def config_file_valid_for_imd(configfile):
    """
    Checks that the config file contains all required parameters for an IMD run.

    Args:
        configfile (str): Name of NAMD config file.

    Returns:
        bool: True if the file is valid, else False.

    """
    if not os.path.exists(configfile):
        print('Error: cannot find {}'.format(configfile))
        return False

    pcoupl = None
    imd = None
    
    with open(configfile, 'r') as f:
        for line in f:
            w = line.split()
            if len(w) > 0:
                if w[0] == 'langevinPiston':
                    pcoupl = w[1]
                if w[0] == 'IMDon':
                    imd = w[1]
    if pcoupl == 'on':
        print('Error: Tios does not support constant pressure simulations')
        print('Please set \"langevinPiston off\" in your .conf file.')
        return False

    if imd != 'yes':
        print("Error: IMD has not been enabled for this simulation.")
        print("Please add \"IMDon   yes\" to your .conf file.")
        return False

    return True

def dt_from_config_file(configfile):
    """
    Extracts the MD timestep from the config file.

    Args:
        configfile (str): Name of the NAMD config file.

    Returns:
        float: The basic MD timestep unit (in picoseconds).

    Raises:
        RuntimeError: if the value of `dt` cannot be determined.

    """
    dt = None
    
    with open(configfile, 'r') as f:
        for line in f:
            w = line.split()
            if len(w) > 0:
                if w[0] == 'timestep':
                    dt = float(w[1])
    if dt is None:
        raise RuntimeError('Error: cannot find MD timestep info in .conf file')
    return dt / 1000 # convert from fs to ps.

def trate_from_config_file(configfile):
    """
    Extracts the IMD sample frequency from the config file.

    Args:
        configfile (str): Name of the NAMD config file.

    Returns:
        float: The IMD sampling frequence (in timesteps).

    Raises:
        RuntimeError: if the value of `trate` cannot be determined.

    """
    trate = None
    
    with open(configfile, 'r') as f:
        for line in f:
            w = line.split()
            if len(w) > 0:
                if w[0] == 'IMDfreq':
                    trate = float(w[1])
    if trate is None:
        raise RuntimeError('Error: cannot find IMD sample rate info in .conf file')
    return trate

def step_from_config_file(configfile):
    """
    Extracts the step number from the config file.

    Args:
        configfile (str): Name of the NAMD config file.

    Returns:
        float: The current time step

    """
    step = None
    
    with open(configfile, 'r') as f:
        for line in f:
            w = line.split()
            if len(w) > 0:
                if w[0] == 'step':
                    dt = int(w[1])
    if step is None:
        raise RuntimeError('Error: cannot find timestep info in .conf file')
    return step

def trajectory_from_config_file(configfile):
    """
    Create an MDTraj Trajectory from the configfile.

    Args:
        configfile (str): Name of the NAMD config file.

    Returns:
        MDTraj.Trajectory: Trajectory object. 

    """
    topfile = None
    
    dir = os.path.dirname(configfile)
    with open(configfile, 'r') as f:
        for line in f:
            w = line.split()
            if len(w) > 0:
                if w[0] == 'coordinates':
                    topfile = w[1]
    if topfile is None:
        raise RuntimeError('Error: cannot find coordinates file info in .conf file')
    trajectory = mdt.load(os.path.join(dir, topfile))
    return trajectory

def step_from_checkpoint_file(cptfile):
    """
    Extracts the current time step from a checkpoint file.

    Args:
        cptfile (str): Name of the checkpoint file.

    Returns:
        int: The final timestep (in units of `dt`).

    Raises:
        RuntimeError: if `cptfile` cannot be found, or the current
            timestep cannot be identified.

    """
    raise NotImplementedError('Not supported for NAMD')

def namd_check_version():
    '''
    Return the version of NAMD available on this resource.

    Returns:
        The NAMD version as a string, or None if it cannot be determined.

    '''
    namd = None
    try:
        output = subprocess.check_output(['which', 'namd2'], stderr=subprocess.STDOUT)
        namd = 'namd2'
    except:
        raise RuntimeError('Cannot find namd2 - is it in your path?')
    try:
        output2 = subprocess.check_output([namd, '-h'], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
            output2 = e.output
    version = 'unknown'
    for line in output2.split(b'\n'):
        words = line.split()
        if len(words) > 0:
            if words[0] == b'Info:' and words[1] == b'NAMD':
                version = words[2].decode()
                break

    return version

def filenames_from_config_file(configfile):
    """
    Parse a NAMD config file and return a list of names of input files.

    Args:
        configfile (str): Name of the configuration file.

    Returns:
        list: List of file names

    """
    result = [configfile]
    dir = os.path.dirname(configfile)
    with open(configfile, 'r') as f:
        for line in f:
            if len(line) > 0:
                words = line.split()
                if len(words) > 1:
                    if words[0] in ['structure', 'coordinates', 'parameters',
                                    'bincoordinates', 'extendedSystem']:
                        result.append(os.path.join(dir, words[1]))
    return result

def apply_checkpoint(inputs, checkpoint):
    """
    Modify the inputs for a NAMD run according to checkpoint data.

    Args:
        inputs (dict): Dictionary of inputs
        checkpoint (dict): NAMD-format checkpoint data

    Returns:
        dict: Modified dictionary of inputs

    """
    configfile = config_file_name(inputs)
    restartfilebase = restart_file_basename(configfile)
    dir = os.path.dirname(restartfilebase)
    for key in checkpoint:        
        with open(os.path.join(dir, key), 'wb') as f:
            f.write(zlib.decompress(checkpoint[key]))

    with open(restartfilebase + '.xsc', 'r') as f:
        for line in f:
            step = line.split()[0]

    with open(configfile, 'r') as f:
        oldconf = f.readlines()
    os.rename(configfile, configfile + '.bak')
    with open(configfile, 'w') as f:
        f.write('bincoordinates     '+ restartfilebase + '.coor\n')
        f.write('binvelocities      '+ restartfilebase + '.vel\n')
        f.write('extendedSystem     '+ restartfilebase + '.xsc\n')
        f.write('firsttimestep      '+ step + '\n')
        for line in oldconf:
            words = line.split()
            if len(words) > 0:
                skip = False
                for keyword in ['binvelocities', 'bincoordinates', 
                                'extendedSystem', 'temperature', 
                                'firsttimestep']:
                    if words[0] == keyword:
                        skip = True
                if not skip:
                    f.write(line)
    return inputs

def config_file_name(inputs):
    """
    Extracts the name of the config file from the inputs dictionary.

    Args:
        inputs (dict): dictionary of inputs

    Returns:
        string: name of config file

    """
    result = None
    for key in inputs:
        if isinstance(inputs[key], str):
            if not inputs[key][0] in '<>+':
                result = inputs[key]
    return result

def restart_file_basename(configfile):
    """
    Returns the basename of the restart files.

    Args:
        configfile (str): configuration file

    Returns:
        str: basename of restart files

    """
    result = None
    dir = os.path.dirname(configfile)
    vars = {}
    with open(configfile, 'r') as f:
        for line in f:
            words = line.split()
            if len(words) > 0:
                if words[0] == 'set':
                    vars['$' + words[1]] = words[2]
                elif words[0] == 'outputName':
                    result = words[1]
                    break
    if result[0] == '$':
        result = vars[result]
    return os.path.join(dir, result) + '.restart'

def step_from_config_file(configfile):
    """
    Returns the start step number set in the config file

    Args:
        configfile (str): config file name

    Returns:
        int: step number

    """
    result = None
    with open(configfile, 'r') as f:
        for line in f:
            if 'step' in line:
                words = line.split()
                if words[0] == 'step':
                    result = int(words[1])
    return result

def inputs_dict_to_filepack(inputs, infiles):
    """
    Convert an inputs dictionary to a filepack - namd style.

    Args:
        inputs (dict): a dictionary of inputs.
        infiles (list): a list of required input files

    Returns:
        list

    """
    for key in inputs:
        if inputs[key] in infiles:
            inputs[key] = os.path.basename(inputs[key])
    filepack = {}
    for file in infiles:
        with open(file, 'rb') as f:
            filepack[os.path.basename(file)] = zlib.compress(f.read())
    return [inputs, filepack]

def filepack_to_inputs_dict(filepack, targetdir='.'):
    """
    Unpack a filepack.

    Args:
        filepack (list): a two element list, first is the inputs dict,
            second is the dictionary of compressed files.
        targetdir (str): optional: directory to uncompress files into.

    Returns:
        dict: The inputs dict.

    """
    inputs = filepack[0]
    filedict = filepack[1]
    for file in filedict:
        with open(os.path.join(targetdir, file), 'wb') as f:
            f.write(zlib.decompress(filedict[file]))

    infiles = [key for key in filedict]
    for key in inputs:
        if inputs[key] in infiles:
            inputs[key] = os.path.join(targetdir, inputs[key]) 
    return inputs
    



import unittest
from tios import communication
import warnings
import sys

class TestTiosDataMethods(unittest.TestCase):

    def setUp(self):
        try:
            self.te = communication.TiosAgent('abc123', new=True)
        except ValueError:
            self.te = communication.TiosAgent('abc123')
        if sys.version_info[0] > 2:
            warnings.simplefilter("ignore", ResourceWarning)

    def test_initialize_new_tios_entry(self):
        self.assertIsInstance(self.te, communication.TiosAgent)

    def test_set_title(self):
        title = 'New Tios Simulation'
        self.te.title = title
        self.te.sync()
        self.assertEqual(title, self.te.title)

    def test_sync(self):
        self.te.status = "Testing"
        self.assertEqual(self.te.updates, {'status' : 'Testing'})
        self.te.sync()
        self.assertEqual(self.te.updates, {})



import unittest
import tempfile
import shutil
import os
from tios import utilities

class TestCheckMethods(unittest.TestCase):

    def setUp(self):
        self.testdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.testdir)

    def test_id_generator(self):
        result1 = utilities.id_generator()
        result2 = utilities.id_generator()
        self.assertNotEqual(result1, result2)

    def test_check_md_code_from_string_with_nothing(self):
        result = utilities.md_code_from_string('')
        self.assertEqual(result, None)

    def test_check_md_code_from_string_with_gmx(self):
        result = utilities.md_code_from_string('mpirun gmx mdrun -deffnm test')
        self.assertEqual(result, 'GROMACS')

    def test_check_md_code_from_string_with_gmx_d(self):
        result = utilities.md_code_from_string('mpirun gmx_d mdrun -deffnm test')
        self.assertEqual(result, 'GROMACS')

    def test_check_md_code_from_string_with_sander(self):
        result = utilities.md_code_from_string('mpirun sander -O -i test.in')
        self.assertEqual(result, 'AMBER')

    def test_check_md_code_from_string_with_pmemd(self):
        result = utilities.md_code_from_string('mpirun pmemd -O -i test.in')
        self.assertEqual(result, 'AMBER')

    def test_check_md_code_from_string_with_pmemd_MPI(self):
        result = utilities.md_code_from_string('mpirun pmemd.MPI -O -i test.in')
        self.assertEqual(result, 'AMBER')

    def test_check_md_code_from_string_with_pmemd_then_gmx(self):
        result = utilities.md_code_from_string('mpirun pmemd -O -i gmx')
        self.assertEqual(result, 'AMBER')

    def test_check_md_code_from_string_with_pmemd_then_gmx_index(self):
        code, result = utilities.md_code_from_string('mpirun pmemd -O -i gmx', index=True)
        self.assertEqual(result, 1)

    def test_check_md_code_from_string_with_namd(self):
        result = utilities.md_code_from_string('mpirun -np 2 namd2 +p4 test.conf')
        self.assertEqual(result, 'NAMD')

    def test_check_command_line_split_gromacs_basic(self):
        preamble, executable, arguments = utilities.command_line_split('mpirun gmx mdrun -deffnm test')
        self.assertTupleEqual((preamble, executable, arguments),(['mpirun'], ['gmx','mdrun'], ['-deffnm', 'test']))

    def test_check_command_line_split_amber_basic(self):
        preamble, executable, arguments = utilities.command_line_split('sander -O -i test.in')
        self.assertTupleEqual((preamble, executable, arguments),([], ['sander'], ['-O', '-i', 'test.in']))

    def test_check_command_line_split_namd_basic(self):
        preamble, executable, arguments = utilities.command_line_split('namd2 test.conf')
        self.assertTupleEqual((preamble, executable, arguments),([], ['namd2'], ['test.conf']))

    def test_check_command_line_split_namd_advanced(self):
        preamble, executable, arguments = utilities.command_line_split('aprun -n 2 namd2 +p8 test.conf')
        self.assertTupleEqual((preamble, executable, arguments),(['aprun', '-n', '2'], ['namd2'], ['+p8', 'test.conf']))

    def test_check_string_to_inputs_dict_gromacs_basic(self):
        inputs = utilities.string_to_inputs_dict('mpirun gmx mdrun -deffnm test -imdwait')
        expected_inputs = {-1 : ['mpirun'], 0 : ['gmx', 'mdrun'],
                           '-s' : 'test.tpr', 
                           '-g' : 'test.log', 
                           '-e' : 'test.edr', 
                           '-o' : 'test.trr', 
                           '-x' : 'test.xtc', 
                           '-cpo' : 'test.cpt', 
                           '-imdwait' : None}
        self.assertDictEqual(expected_inputs, inputs)

    def test_check_string_to_inputs_dict_amber_basic(self):
        inputs = utilities.string_to_inputs_dict('pmemd -O -i test.in -o test.out')
        expected_inputs = {-1 : [], 0 : ['pmemd'], '-O' : None, '-i' : 'test.in', '-o' : 'test.out'}
        self.assertDictEqual(expected_inputs, inputs)

    def test_check_string_to_inputs_dict_namd_basic(self):
        inputs = utilities.string_to_inputs_dict('namd2 +p4 test.conf')
        expected_inputs = {-1 : [], 0 : ['namd2'], 1 : '+p4', 2 : 'test.conf'}
        self.assertDictEqual(expected_inputs, inputs)
    '''
    def test_inputs_dict_to_string_gromacs(self):
        expected_command = 'mpirun gmx mdrun -deffnm test -imdwait'
        inputs = utilities.string_to_inputs_dict(expected_command)
        command = utilities.inputs_dict_to_string(inputs)
        self.assertEqual(expected_command, command)
    '''

    def test_inputs_dict_to_string_amber(self):
        expected_command = 'pmemd.MPI -O -i test.in -o test.out'
        inputs = utilities.string_to_inputs_dict(expected_command)
        command = utilities.inputs_dict_to_string(inputs)
        options = ['pmemd.MPI -O -i test.in -o test.out',
                   'pmemd.MPI -O -o test.out -i test.in',
                   'pmemd.MPI -i test.in -O -o test.out',
                   'pmemd.MPI -i test.in -o test.out -O',
                   'pmemd.MPI -o test.out -O -i test.in',
                   'pmemd.MPI -o test.out -i test.in -O']
        self.assertTrue(command in options)

    def test_inputs_dict_to_string_namd(self):
        expected_command = 'namd2 +p4 test.conf' 
        inputs = utilities.string_to_inputs_dict(expected_command)
        command = utilities.inputs_dict_to_string(inputs)
        self.assertEqual(command, expected_command)

    def test_inputs_dict_to_filepack_gromacs(self):
        command = 'gmx mdrun -s test/examples/bpti.tpr -imdwait'
        inputs = utilities.string_to_inputs_dict(command)
        filepack = utilities.inputs_dict_to_filepack(inputs)
        self.assertIsInstance(filepack, dict)

    def test_filepack_to_inputs_dict(self):
        command = 'gmx mdrun -s test/examples/bpti.tpr -imdwait'
        inputs = utilities.string_to_inputs_dict(command)
        filepack = utilities.inputs_dict_to_filepack(inputs)
        new_inputs = utilities.filepack_to_inputs_dict(filepack, targetdir=self.testdir)
        self.assertEqual(new_inputs['-s'], '{}/bpti.tpr'.format(self.testdir))
        self.assertTrue(os.path.exists(new_inputs['-s']))




import unittest
from tios import imd_sim, communication

class TestDummyConnectionMethods(unittest.TestCase):

    def test_initialize_new_dummy_connection(self):
        connection = communication.select('Dummy', 'abc123')
        self.assertIsInstance(connection, communication.DummyConnection)

    def test_has_entry(self):
        connection = communication.DummyConnection('abc123')
        self.assertFalse(connection.has_entry())
        connection.add_entry()
        self.assertTrue(connection.has_entry())

    def test_set_title(self):
        title = 'New Tios Simulation'
        connection = communication.DummyConnection('abc123')
        update = {'title': title}
        connection.put(update)
        self.assertEqual(title, connection.get('title'))




from tios import gromacs_utilities, namd_utilities

import subprocess
import os
import zlib
import string
import random

#Taken with thanks from 
#https://stackoverflow.com/questions/18499497/how-to-process-sigterm-signal-gracefully
import signal
class GracefulKiller(object):
    kill_now = False
    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        signal.siginterrupt(signal.SIGINT, False)
        signal.siginterrupt(signal.SIGTERM, False)

    def exit_gracefully(self, signum, frame):
        print('killer received signal {}'.format(signum))
        self.kill_now = True

def id_generator(size=6, chars=string.ascii_lowercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))

def command_line_split(command):
    '''
    Splits a command line into preamble, executable, and arguments.
    '''
    preamble = []
    executable = []
    arguments = []
    code, indx = md_code_from_string(command, index=True)
    words = command.split()
    if code == 'GROMACS':
        preamble = words[0:indx]
        executable = words[indx:indx + 2]
        arguments = words[indx + 2:]
    elif code is not None:
        preamble = words[0:indx]
        executable = [words[indx]]
        arguments = words[indx + 1:]
    return preamble, executable, arguments 
        
def string_to_inputs_dict(command):
    '''
    Given a command line string, converts it into an inputs dictionary
    according to ExTASY wrappers philosophy.
    '''
    inputs = {}
    preamble, executable, arguments = command_line_split(command)
    inputs[-1] = preamble
    inputs[0] = executable
    n_pos = 0 # number of positional arguments
    n_args = len(arguments)
    iarg = 0
    while iarg < n_args:
        if not arguments[iarg][0] in '-<>':
            n_pos += 1
            inputs[n_pos] = arguments[iarg]
            iarg += 1
        else:
            if iarg + 1 < n_args:
                if arguments[iarg + 1][0] in '-<>':
                    inputs[arguments[iarg]] = None
                    iarg += 1
                else:
                    inputs[arguments[iarg]] = arguments[iarg + 1]
                    iarg += 2
            else:
                inputs[arguments[iarg]] = None
                iarg += 1
    md_code = md_code_from_string(command)
    if md_code == 'GROMACS':
        inputs = gromacs_utilities.complete_inputs(inputs)
    return inputs
               
def inputs_dict_to_string(inputs):
    '''
    Given a dictionary of inputs, return a command line string.
    '''
    # preamble:
    command = " ".join(p for p in inputs[-1])
    if len(command) > 0:
        command += " "

    # executable:
    command += " ".join(e for e in inputs[0])

    # positional arguments:
    max_pos_arg = 0
    for key in inputs:
        if isinstance(key, int):
            if key > max_pos_arg:
                max_pos_arg = key
    for pos_arg in range(1, max_pos_arg + 1):
        command += " " + inputs[pos_arg]

    # other arguments:
    for key in inputs:
        if not isinstance(key, int):
            command += " " + key 
            if isinstance(inputs[key], tuple): # for filepacks
                command += " " + inputs[key][0]
            elif inputs[key] is not None:
                command += " " + inputs[key]
    return command
    
def inputs_dict_to_filepack(inputs):
    '''
    Given a dictionary of inputs, returns a filepack.
    A filepack is like an inputs dictionary, except that
    where the argument to a key in inputs is the name of
    an existing file, in the filepack dictionary it is a
    (filename, contents) tuple, where contents is zlib
    compressed.
    '''
    command = inputs_dict_to_string(inputs)
    code = md_code_from_string(command)
    if code == 'GROMACS':
        input_keys = ['-s', '-cpi']
    elif code == 'NAMD':
        infiles = filenames_in_inputs_dict(inputs)
        return namd_utilities.inputs_dict_to_filepack(inputs, infiles)
    else:
        raise NotImplementedError('MD code {} is not supported'.format(code))
    filepack = inputs.copy()
    for key in filepack:
        if isinstance(filepack[key], str):
            if key in input_keys and os.path.exists(filepack[key]):
                filename = filepack[key]
                with open(filename, 'rb') as f:
                    filepack[key] = (filename, zlib.compress(f.read()))
    return filepack

def filenames_in_inputs_dict(inputs):
    '''
    Parses a dictionary of inputs and returns a list of all the filenames
    '''
    command = inputs_dict_to_string(inputs)
    code = md_code_from_string(command)
    result = []
    if code == "GROMACS":
        for key in inputs:
            if key in ['-s', '-cpi', '-o', '-cpo', '-g', '-x', '-c', '-e',
                       '-dhdl', '-field', '-table', '-tabletf', '-tablep',
                       '-rerun', '-tpi', '-tpid', '-ei', '-eo', '-devout',
                       '-runav', '-px', '-pf', '-ro', '-ra', '-rs', '-rt',
                       '-mtx', '-dn', '-multidir', '-mp', '-mn', '-if',
                       '-swap']:
                result.append(inputs[key])
        return result
    elif code == 'NAMD':
        for key in inputs:
            value = inputs[key]
            if (not isinstance(value, list)) and value[0] != '+':
                result += namd_utilities.filenames_from_config_file(inputs[key])
        return result
    else:
        raise NotImplementedError('Sorry - MD code {} not supported'.format(code))
        

def filepack_to_inputs_dict(filepack, targetdir=None):
    '''
    Given a filepack, unpacks it  and returns an inputs dictionary
    '''
    if isinstance(filepack, list):
        return namd_utilities.filepack_to_inputs_dict(filepack, targetdir=targetdir)
    inputs = filepack.copy()
    filenames = filenames_in_inputs_dict(inputs)
    for key in inputs:
        if inputs[key] in filenames:
            if isinstance(inputs[key], tuple):
                filename = inputs[key][0]
                if targetdir is not None:
                    filename = os.path.join(targetdir, os.path.basename(filename))
                with open(filename, 'wb') as f:
                    f.write(zlib.decompress(inputs[key][1]))
                inputs[key] = filename
            else:
                filename = inputs[key]
                if targetdir is not None:
                    filename = os.path.join(targetdir, os.path.basename(filename))
                inputs[key] = filename
    return inputs

def md_code_from_string(command, index=False):
    '''
    Given a command line string, identify/guess the relevant MD code.
    Returns a string, or None if it cannot identify a likely code.

    If index is not none, then the index of the identificed keyword is
    also ruturned with the code name as a (codename, index) tuple.
    '''
    code = None
    words = command.split()
    # The heuristic is that the leftmost key word takes precedence. This
    # should help to overcome pathological cases such as 
    #    'gmx mdrun -deffnm sander' 
    # being misinterpreted.
    leftmost = len(words)
    codes = {'gmx' : 'GROMACS',
            'gmx_d' : 'GROMACS',
            'sander' : 'AMBER',
            'pmemd' : 'AMBER',
            'pmemd.MPI' : 'AMBER',
            'namd2' : 'NAMD',
            'namd' : 'NAMD' }
    for key in codes:
        for word in words:
            if key in word:
                if words.index(word) < leftmost:
                    leftmost = words.index(word)
                    code = codes[key]
    if not index:
        return code
    else:
        return code, leftmost
    
    
def installed_version(code):
    '''
    Check that the required MD code is available, and if so, what version it is
    '''
    if code == 'GROMACS':
        return gromacs_utilities.gromacs_check_version()
    elif code == 'NAMD':
        return namd_utilities.namd_check_version()
    else:
        raise NotImplementedError('{} is not currently supported'.format(code))




import unittest
from tios import communication 

class TestMongoConnectionMethods(unittest.TestCase):

    def test_initialize_new_mongo_connection(self):
        connection = communication.select('Mongo', 'abc123')
        self.assertIsInstance(connection, communication.MongoConnection)

    def test_has_entry(self):
        connection = communication.MongoConnection('abc123')
        while connection.has_entry():
            connection.delete_entry()
        self.assertFalse(connection.has_entry())
        connection.add_entry()
        self.assertTrue(connection.has_entry())

    def test_set_title(self):
        title = 'New Tios Simulation'
        connection = communication.MongoConnection('abc123')
        update = {'title': title}
        connection.put(update)
        self.assertEqual(title, connection.get('title'))




from __future__ import print_function
from bson.binary import Binary
import pickle
import zlib
import sys
import pymongo

from tios import environments
'''
The Tios communication module.

This module provides the generalised API for communication between the
simulation and the database.
'''

python3 = sys.version_info[0] == 3
def select(protocol, id):
    """
    Select the connection protocol.

    Args:
        protocol (str): The protocol to use. Currently only the options
            "Dummy" and "Mongo" are recognised.
        id (str): The ID of the entry

    Returns:
        connection (class TiosConnection): an instance of a `TiosConnection` 
            with "get", "put" methods, etc.
    """
    if protocol == 'Dummy':
        return DummyConnection(id)
    elif protocol == 'Mongo':
        return MongoConnection(id)
    else:
        raise ValueError('Error: protocol {} is not supported'.format(protocol))

class TiosConnection(object):
    """
    Base class for the database connection layer in Tios.
    """

    def __init__(self, id):
        self.id = id
        self.access = 'rw'

    def put(self, updates):
        """
        Update entry with data from the dictionary `updates`.

        Args:
            updates (dict): Dictionary of keys and values to be updated

        """
        raise NotImplementedError

    def get(self, key):
        """
        Returns the item with the given `key`.

        Args:
            key (str): Key of the Tios entry item to retrieve

        Returns:
            value: The value of the `key`, in whatever type is appropriate

        """
        raise NotImplementedError

    def has_entry(self):
        """
        Tests if the database has entry.

        Returns:
            is_present (bool): ``True`` if the id is found, otherwise ``False``
        """
        raise NotImplementedError

    def add_entry(self):
        """
        Create a new database record 

        """
        raise NotImplementedError

    def delete_entry(self):
        """
        Remove record from the database

        """
        raise NotImplementedError


class DummyConnection(TiosConnection):
    """
    A dummy connection object for testing purposes.
    """
    def __init__(self, id):
        self.id = None
        self._id = id
        self.data = {}
        self.access = 'rw'

    def put(self, updates):
        for key in updates:
            self.data[key] = updates[key]

    def get(self, key):
        return self.data[key]

    def has_entry(self):
        return self.id != None

    def add_entry(self):
        self.id = self._id

    def delete_entry(self):
        self.id = None

class MongoConnection(TiosConnection):
    """
    A Mongo-based connection object.
    """
    def __init__(self, id):

        self.id = id
        self.access = 'r'
        self.env = environments.load_mongo_env()
        self.collection = self.env['COLLECTION']
        if self.has_entry():
            self.access = 'rw'
        else:
            coll_names = [c for c in self.collection.database.collection_names()]
            for c in coll_names:
                self.collection = self.collection.database[c]
                if self.has_entry():
                    self.access = 'r'
                    break
               
            if not self.has_entry():
                self.collection = self.env['COLLECTION']
                self.access = 'rw'

    def put(self, updates):
        if not 'w' in self.access:
            raise IOError('Error: you do not have write access to this entry')
        success = False
        while not success:
            try:
                result = self.collection.update_one(
                    {"id" : self.id},
                    {
                        "$set" : updates,
                        "$currentDate" : {"last_update" : True}
                    }
                )
                success = True
            except pymongo.errors.AutoReconnect:
                pass
        if result.matched_count != 1:
            print('Warning: TiosConnection.put(): update failed')

    def get(self, key):
        success = False
        while not success:
            try:
                result = self.collection.find_one(
                    {"id" : self.id},
                    projection=[key]
                )
                success = True
            except pymongo.errors.AutoReconnect:
                pass
        return result[key]

    def has_entry(self):
        success = False
        while not success:
            try:
                result = self.collection.find_one({'id': self.id}, 
                                                projection=['id']) is not None
                success = True
            except pymongo.errors.AutoReconnect:
                pass
        return result

    def add_entry(self):
        if not 'w' in self.access:
            raise IOError('Error: you do not have write access to this entry')
        data = {"id" : self.id}
        success = False
        while not success:
            try:
                self.collection.insert_one(data)
                success = True
            except pymongo.errors.AutoReconnect:
                pass

    def delete_entry(self):
        if not 'w' in self.access:
            raise IOError('Error: yYou do not have write access to this entry')
        success = False
        while not success:
            try:
                result = self.collection.delete_one({"id" : self.id})
                success = True
            except pymongo.errors.AutoReconnect:
                pass

class TiosAgent(object):
    """
    Class that provides a local interface to the tios data model.
    """
    def __init__(self, id, protocol='Mongo', new=False):
        """
        Create a local interface to an entry in the remote Tios datastore.

        Args:
            id (str): Tios entry id/code
            protocol (str, optional): The communication protocol. 
                Default is "Mongo".
            new (bool, optional): If True a new entry is created.

        Returns:
            agent (class TiosAgent): an object with properties to set and get 
                each key in the associated database entry.

        Raises:
            ValueError: If `new` is False and `id` does not already exist, or
                if `new` is True and `id` already exists.

        """
        self.data_connection = select(protocol, id)
        if new:
            if self.data_connection.has_entry():
                raise ValueError('Error: this entry already exists')
            else:
                self.data_connection.add_entry()
        else:
            if not self.data_connection.has_entry():
                raise ValueError('Error: this entry does not exist')
        self.id = id
        self.updates = {}
        self._title = None
        self._timepoint = None
        self._last_update = None
        self._status = None
        self._message = None
        self._md_code = None
        self._md_version = None
        self._host = None
        self._username = None
        self._trate = None
        self._frame_rate = None
        self._filepack = None
        self._checkpoint = None
        self._topology = None
        self._selection = None
        self._box = None
        self._xyzsel = None
        self._xyzunsel = None
        self._monitors = None

    def sync(self):
        """
        Synchronise updates to the local data object with the remote database
        """
        self.data_connection.put(self.updates)
        self.updates = {}

    @property
    def title(self):
        """
        String giving the user-supplied title for this Tios simulation.
        """
        self._title = self.data_connection.get('title')
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        self.updates['title'] = value

    @property
    def timepoint(self):
        """
        Float giving the age of the simulation (in picoseconds).
        """
        self._timepoint = self.data_connection.get('timepoint')
        return self._timepoint

    @timepoint.setter
    def timepoint(self, value):
        self._timepoint = value
        self.updates['timepoint'] = value

    @property
    def last_update(self):
        """
        Time-stamp of the last update to the remote database.
        """
        self._last_update = self.data_connection.get('last_update')
        return self._last_update

    @last_update.setter
    def last_update(self, value):
        self._last_update = value
        self.updates['last_update'] = value

    @property
    def status(self):
        """
        String giving the current status of the Tios simulation.
        """
        self._status = self.data_connection.get('status')
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        self.updates['status'] = value

    @property
    def message(self):
        """
        Dictionary to send messages to and from a running simulation.
        """
        self._message = self.data_connection.get('message')
        return self._message

    @message.setter
    def message(self, value):
        self._message = value
        self.updates['message'] = value

    @property
    def md_code(self):
        """
        String defining the MD code: "GROMACS", "NAMD", etc.
        """
        self._md_code = self.data_connection.get('md_code')
        return self._md_code

    @md_code.setter
    def md_code(self, value):
        self._md_code = value
        self.updates['md_code'] = value

    @property
    def md_version(self):
        """
        String defining the version of the MD code in current/last use.
        """
        self._md_version = self.data_connection.get('md_version')
        return self._md_version

    @md_version.setter
    def md_version(self, value):
        self._md_version = value
        self.updates['md_version'] = value

    @property
    def host(self):
        """
        String giving the current or most recent hostname for the simulation.
        """
        self._host = self.data_connection.get('host')
        return self._host

    @host.setter
    def host(self, value):
        self._host = value
        self.updates['host'] = value

    @property
    def username(self):
        """
        String with the current/most recent username running the simulation.
        """
        self._username = self.data_connection.get('username')
        return self._username

    @username.setter
    def username(self, value):
        self._username = value
        self.updates['username'] = value

    @property
    def trate(self):
        """
        Float giving the interval between IMD samples (in picoseconds).
        """
        self._trate = self.data_connection.get('trate')
        return self._trate

    @trate.setter
    def trate(self, value):
        self._trate = value
        self.updates['trate'] = value

    @property
    def frame_rate(self):
        """
        Float giving the current/most recent number of IMD frames per minute.
        """
        self._frame_rate = self.data_connection.get('frame_rate')
        return self._frame_rate

    @frame_rate.setter
    def frame_rate(self, value):
        self._frame_rate = value
        self.updates['frame_rate'] = value

    @property
    def filepack(self):
        """
        A dictionary of ``{filename : filecontents}`` pairs of input files.
        """
        self._filepack = pickle.loads(self.data_connection.get('filepack'))
        return self._filepack

    @filepack.setter
    def filepack(self, value):
        self._filepack = value
        self.updates['filepack'] = Binary(pickle.dumps(value))

    @property
    def checkpoint(self):
        """
        A MD-code specific checkpoint object.
        """
        self._checkpoint = pickle.loads(self.data_connection.get('checkpoint'))
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, value):
        self._checkpoint = value
        self.updates['checkpoint'] = Binary(pickle.dumps(value))

    @property
    def topology(self):
        """
        An MDTraj.Topology for the system.
        """
        self._topology = pickle.loads(zlib.decompress(self.data_connection.get('topology')))
        return self._topology

    @topology.setter
    def topology(self, value):
        self._topology = value
        self.updates['topology'] = Binary(zlib.compress(pickle.dumps(value)))

    @property
    def selection(self):
        """
        A list defining the indices of all "selected" atoms.
        """
        self._selection = self.data_connection.get('selection')
        return self._selection

    @selection.setter
    def selection(self, value):
        self._selection = value
        self.updates['selection'] = value

    @property
    def box(self):
        """
        A numpy array of floats with the periodic box information.
        """
        result = self.data_connection.get('box')
        if python3:
            self._box = pickle.loads(result, encoding='latin1')
        else:
            self._box = pickle.loads(result)
        return self._box

    @box.setter
    def box(self, value):
        self._box = value
        self.updates['box'] = Binary(pickle.dumps(value))

    @property
    def xyzsel(self):
        """
        A numpy array of floats with the coordinates of the selected subset
        """ 
        if python3:
            self._xyzsel = pickle.loads(zlib.decompress(self.data_connection.get('xyzsel')), encoding='latin1')
        else:
            self._xyzsel = pickle.loads(zlib.decompress(self.data_connection.get('xyzsel')))
        return self._xyzsel

    @xyzsel.setter
    def xyzsel(self, value):
        self._xyzsel = value
        self.updates['xyzsel'] = Binary(zlib.compress(pickle.dumps(value)))

    @property
    def xyzunsel(self):
        """
        A numpy array of floats with the coordinates of non-selected atoms.
        """
        if python3:
            self._xyzunsel = pickle.loads(zlib.decompress(self.data_connection.get('xyzunsel')), encoding='latin1')
        else:
            self._xyzunsel = pickle.loads(zlib.decompress(self.data_connection.get('xyzunsel')))
        return self._xyzunsel

    @xyzunsel.setter
    def xyzunsel(self, value):
        self._xyzunsel = value
        self.updates['xyzunsel'] = Binary(zlib.compress(pickle.dumps(value)))

    @property
    def monitors(self):
        """
        A dictionary of simulation metrics
        """
        if python3:
            self._monitors = pickle.loads(self.data_connection.get('monitors'), encoding='latin1')
        else:
            self._monitors = pickle.loads(self.data_connection.get('monitors'))
        return self._monitors

    @monitors.setter
    def monitors(self, value):
        self._monitors = value
        self.updates['monitors'] = Binary(pickle.dumps(value))




'''
A simple and generic control interface for MD simulation jobs under IMD control.

An instance of a simulation job is created from a string of the sort that
might have been a command line, e.g.:

    sim = from_command('gmx mdrun -deffnm test')

An exception will be raised if the job is not a valid IMD job.

The simulation is started only when the start method is called:

    sim.start()

The status of the simulation can be checked using the is_running method:

    is_running = (sim.is_running() == True)

The simulation is advanced to the next IMD timestep by the step method:

    sim.step()

The IMD sample interval can be changed during a simulation:

    sim.set_trate(trate)

The simulation can be stopped using the stop method:

    sim.stop()

Once stopped, the standard output and standard error from the job are
available:

    print sim.stdout
    print sim.stderr

An MDTraj Trajectory object with a single frame - the current coordinates -
is available:

    topology = sim.trajectory.topology
    crds = sim.trajectory.xyz

The current energy components, etc. are available in the energies attribute 
(a dictionary). The contents of this are defined by the IMD protocol, the
keys are: 'tstep', 'T', 'Etot', 'Epot', 'Evdw', 'Eelec', 'Ebond', 'Eangle', 
'Edihe', 'Eimpr'.

Note that sim.energies['tstep'] returns the current timestep number; the
'age' of the simulation, in picoseconds, is available in the 'timepoint' attribute:

    age = sim.timepoint

The underlying MD code may feature checkpointing. If so, then the attribute
new_checkpoint will be True if an updated checkpoint is available. It is
returned by the get_checkpoint() method, which then resets new_checkpoint to
False. Exactly what is returned by get_checkpoint() is MD-code dependent, but 
the contents can be passed back to a new instatiation of the run as an
optional argument:

    if sim.has_checkpoint:
        new_chk = sim.get_checkpoint()
    sim.stop()
    ...
    restarted_sim = from_command('gmx mdrun -deffnm test', checkpoint=new_chk)

'''
from tios import utilities, gromacs_utilities, namd_utilities, imd

import os
import subprocess
import time
import glob
import socket
import zlib
import numpy as np

def from_command(command_line, checkpoint=None, trate=1.0):
    """
    Create an instance of an IMDJob from a command line.

    Optional extra (code-specific) checkpoint data can be provided. The IMD 
    sampling rate can also be set.

    Args:
        command_line (str): A command line that would run the job interactively.
        checkpoint: optional checkpoint data of code-specific type.
        trate (float, optional): IMD sampling inteval (in picoseconds).

    Returns:
        IMDJob: An instance of an IMD-controlled MD job.

    Raises:
        NotImplementedError: If the MD code is not supported.

    """
    md_code = utilities.md_code_from_string(command_line)
    version = utilities.installed_version(md_code)
    if version is not None:
        if md_code == 'GROMACS':
            return GromacsIMDJob(command_line, checkpoint=checkpoint, trate=trate)
        elif md_code == 'NAMD':
            return NAMDIMDJob(command_line, checkpoint=checkpoint, trate=trate)
        else:
            raise NotImplementedError('Error - this is a {} job but this code is currently not supported by tios.'.format(md_code))
    else:
        raise NotImplementedError('Error - this is a {} job but this code is not available on this resource'.format(md_code))

class IMDJob(object):
    '''
    Base class for IMD-controlled MD jobs.
    MD-code specific subclasses are based on this.
    '''
    def __init__(self, command_line, checkpoint=None, trate=1.0):
        """
        Initialises a new IMD controlled simulation job. 

    Args:
        command_line (str): A command line that would run the job interactively.
        checkpoint: optional checkpoint data of code-specific type.
        trate (float, optional): IMD sampling inteval (in picoseconds).

    Attributes:
        status (str): Status of the simulation.
        timepoint (float): Age of the simulation.
        trajectory (MDTraj.Trajectory): Single-frame trajectory object.
        energies (dict): IMD energy components.
        stdout (str): Contents of STDOUT from running job.
        stderr (str): Contents of STDERR from running job.
        new_checkpoint (bool): True if a new checkpoint file is available.

        """

        self.status = "Unknown" 
        self.trajectory = None 
        self.energies = None 
        self.timepoint = 0.0 
        self._stdout = open('STDOUT', 'wb')
        self.stdout = ""
        self.stderr = ""
        self.new_checkpoint = False
        self._subprocess = None # subprocess running MD code
        self._socket = None # socket being used by IMD protocol
        self._dt = 0.001 # simulation timestep

    def start(self):
        """
        Launch the job.
        """
        raise NotImplementedError

    def is_running(self):
        """
        Returns True if the simulation appears to be running.
        """
        self._subprocess.poll()
        if self._subprocess.returncode is None:
            self.status = 'Running'
            return True
        else:
            self.status = 'Stopped'
            stdout, stderr = self._subprocess.communicate()
            if stdout is not None:
                self.stdout += stdout.decode()
            if stderr is not None:
                self.stderr += stderr.decode()
        return False

    def stop(self):
        """
        Stops the IMD simulation.
        """
        imd.imd_kill(self._socket)
        self._socket.close()
        try:
            self._subprocess.terminate()
        except OSError:
            pass
        self.status = 'Stopped'
        time.sleep(2)
        self._update_checkpoint()

    def step(self):
        """
        Move the IMD simulation forward to the next data collection timepoint.

        Also updates the current data attributes.
        """
        new_data = False
        while not new_data:
            typ, length = imd.imd_recv_header(self._socket)
        
            if typ == imd.IMD_ENERGIES:
                self.energies = imd.imd_recv_energies(self._socket)
                self.timepoint = self.energies['tstep'] * self._dt
            elif typ == imd.IMD_FCOORDS:
                n_coords = length
                coords = imd.imd_recv_fcoords(self._socket, n_coords) 
                self.trajectory.xyz = np.array(coords).reshape((-1, 3)) * 0.1
                new_data = True
        self._update_checkpoint()

    def _update_checkpoint(self):
        """
        Ensure the latest checkpoint data is available for get_checkpoint()

        if successful, sets the new_checkpoint attribute to True
        """
        raise NotImplementedError

    def get_checkpoint(self):
        """
        Return the latest checkpoint data.

        This can be in what ever format suits the _apply_checkpoint() method.
        Once called, sets the new_checkpoint attribute to False.
        """
        raise NotImplementedError
 
class GromacsIMDJob(IMDJob):
    def __init__(self, command_line, checkpoint=None, trate=1.0):

        inputs = utilities.string_to_inputs_dict(command_line)
        inputs = gromacs_utilities.complete_inputs(inputs)
        inputs = gromacs_utilities.complete_inputs_for_imd(inputs)

        self._has_checkpoint = False
        self.new_checkpoint = False
        self._last_checkpoint_time = time.time()
        self.stdout = ''
        self.stderr = ''
        self.logfile = 'STDOUT'
        self._stdout = open(self.logfile, 'wb')
        self.timepoint = 0.0

        if checkpoint is not None:
            inputs = gromacs_utilities.apply_checkpoint(inputs, checkpoint)
        if '-s' in inputs:
            tpr_file = inputs['-s']
        else:
            self.stderr += 'Error: cannot identify the GROMACS tpr file for this job\n'
            self.status='Failed'
            raise RuntimeError('Error: cannot identify the GROMACS tpr file for this job')
        if not gromacs_utilities.tpr_file_valid_for_imd(tpr_file):
            self.stderr += 'Error: {} does not define a valid IMD job\n'.format(tpr_file)
            self.status='Failed'
            raise RuntimeError('Error: {} does not define a valid IMD job\n'.format(tpr_file))

        self._dt = gromacs_utilities.dt_from_tpr_file(tpr_file)
        self._trate = trate
        if '-cpi' in inputs:
            step = gromacs_utilities.step_from_checkpoint_file(inputs['-cpi'])
            if step is not None:
                self.timepoint = step * self._dt

        self.trajectory = gromacs_utilities.trajectory_from_tpr_file(tpr_file)
        self._inputs = inputs
        self.status = 'Ready'

    def _update_checkpoint(self):
        if os.path.exists(self._inputs['-cpo']):
            self._has_checkpoint = True
            self._checkpoint_mtime = os.path.getmtime(self._inputs['-cpo'])
            if self._last_checkpoint_time < self._checkpoint_mtime:
                self.new_checkpoint = True
                self._last_checkpoint_time = self._checkpoint_mtime

    def get_checkpoint(self):
        if self._has_checkpoint:
            with open(self._inputs['-cpo'], 'rb') as f:
                result = zlib.compress(f.read())
            self.new_checkpoint = False
            return result

    def start(self):
        try:
            command = utilities.inputs_dict_to_string(self._inputs)
            self._subprocess = subprocess.Popen(command.split(),
                                        stderr=subprocess.STDOUT,
                                        stdout=self._stdout)
        except OSError as e:
            self.status = 'Failed'
            self.stderr += e.strerror
            raise
        time.sleep(5)

        if not self.is_running():
            self.stdout += "Job terminated prematurely\n"
            self.status = 'Failed'
            raise RuntimeError(self.stdout)

        f = open(self.logfile, 'rb')
        output = f.readline()
        while not b'Listening' in output or b'Fatal' in output:
            if self.is_running():
                output = f.readline()
                self.stdout += output.decode()
            else:
                raise RuntimeError('Error - job failed - check log file {}\n'.format(logfile))
        if b'Fatal' in output:
            self.stop()
            raise RuntimeError('Error - job failed - check log file {}.log\n'.format(logfile))
        f.close()
        words = output.split()
        indx = words.index(b'Listening')
        self._port = int(words[indx + 6][:-1])

        logfile = self._inputs['-g']
        logroot = os.path.splitext(logfile)[0]
        real_logfiles = glob.glob(logroot + '*.log')
        if len(real_logfiles) == 0:
            raise RuntimeError('Cannot find a logfile like {}*.log'.format(logroot))
        self._inputs['-g'] = real_logfiles[-1]
        logfile = real_logfiles[-1]

        with open(logfile, 'r') as f:
            output = f.readline()
            while not 'Host:' in output:
                output = f.readline()
        words = output.split()
        indx = words.index('Host:')
        self._hostname = words[indx + 1]
        #self._hostname = 'localhost'
        try:
            self._socket = imd.imd_connect(self._hostname, self._port)
        except socket.gaierror:
            self._hostname = 'localhost'
        except socket.error:
            self._hostname = 'localhost'
        self._socket = imd.imd_connect(self._hostname, self._port)
        imdcheck = imd.imd_recv_handshake(self._socket)
        if imdcheck == -1:
            self.stop()
            raise RuntimeError('Error: IMD handshake failed')
        elif imdcheck == 1:
            self.stop()
            raise RuntimeError('Error: IMD handshake detects endianness problem')
            return
            
        imd.imd_trate(self._socket, self._trate/self._dt)
        self.status = 'Running'

class NAMDIMDJob(IMDJob):
    def __init__(self, command_line, checkpoint=None, trate=1.0):

        inputs = utilities.string_to_inputs_dict(command_line)

        self._has_checkpoint = False
        self.new_checkpoint = False
        self._last_checkpoint_time = time.time()
        self.stdout = ''
        self.stderr = ''
        self.logfile = 'STDOUT'
        self._stdout = open(self.logfile, 'wb')
        self.timepoint = 0.0

        if checkpoint is not None:
            inputs = namd_utilities.apply_checkpoint(inputs, checkpoint)
        configfile = namd_utilities.config_file_name(inputs)
        if not namd_utilities.config_file_valid_for_imd(configfile):
            self.stderr += 'Error: {} does not define a valid IMD job\n'.format(configfile)
            self.status='Failed'
            raise RuntimeError('Error: {} does not define a valid IMD job\n'.format(configfile))

        self._dt = namd_utilities.dt_from_config_file(configfile)
        self._trate = namd_utilities.trate_from_config_file(configfile) * self._dt
        step = namd_utilities.step_from_config_file(configfile)
        if step is not None:
            self.timepoint = step * self._dt

        self.trajectory = namd_utilities.trajectory_from_config_file(configfile)
        self._inputs = inputs
        self._configfile = configfile
        self.status = 'Ready'

    def _update_checkpoint(self):
        xscfile = namd_utilities.restart_file_basename(self._configfile) + '.xsc'
        if os.path.exists(xscfile):
            self._has_checkpoint = True
            self._checkpoint_mtime = os.path.getmtime(xscfile)
            if self._last_checkpoint_time < self._checkpoint_mtime:
                self.new_checkpoint = True
                self._last_checkpoint_time = self._checkpoint_mtime

    def get_checkpoint(self):
        result = {}
        if self._has_checkpoint:
            basename = namd_utilities.restart_file_basename(self._configfile)
            for ext in ['.coor', '.vel', '.xsc']:
                with open(basename + ext, 'rb') as f:
                    result[basename + ext] = zlib.compress(f.read())
            self.new_checkpoint = False
        return result

    def start(self):
        try:
            command = utilities.inputs_dict_to_string(self._inputs)
            self._subprocess = subprocess.Popen(command.split(),
                                        stderr=subprocess.STDOUT,
                                        stdout=self._stdout)
        except OSError as e:
            self.status = 'Failed'
            self.stderr += e.strerror
            raise
        time.sleep(5)

        if not self.is_running():
            self.stdout += "Job terminated prematurely\n"
            self.status = 'Failed'
            raise RuntimeError(self.stdout)

        f = open(self.logfile, 'r')
        output = f.readline()
        while not 'Info: 1 NAMD' in output or 'Fatal' in output:
            if self.is_running():
                output = f.readline()
                self.stdout += output
            else:
                raise RuntimeError('Error - job failed - check log file {}\n'.format(self.logfile))
        if 'Fatal' in output:
            self.stop()
            raise RuntimeError('Error - job failed')
        words = output.split()
        self._hostname = words[6]

        while not 'listening' in output or 'Fatal' in output:
            if self.is_running():
                output = f.readline()
                self.stdout += output
            else:
                raise RuntimeError('Error - job failed')
        if 'Fatal' in output:
            self.stop()
            raise RuntimeError('Error - job failed')
        words = output.split()
        indx = words.index('listening')
        self._port = int(words[indx + 3][:-1])

        try:
            self._socket = imd.imd_connect(self._hostname, self._port)
        except socket.gaierror:
            self._hostname = 'localhost'
        self._socket = imd.imd_connect(self._hostname, self._port)
        imdcheck = imd.imd_recv_handshake(self._socket)
        if imdcheck == -1:
            self.stop()
            raise RuntimeError('Error: IMD handshake failed')
        elif imdcheck == 1:
            self.stop()
            raise RuntimeError('Error: IMD handshake detects endianness problem')
            return
            
        imd.imd_trate(self._socket, self._trate/self._dt)
        self.status = 'Running'
        f.close()



import sys
import datetime
from tios import environments

env = environments.load_mongo_env(administrator=True)
db = env['COLLECTION'].database
db_name = env['DB_NAME']

def add_guest(password):
    '''
    Create the Tios guest account.
    '''
    now = datetime.datetime.utcnow()

    guest_privileges = {'resource':{'db':db_name, 'collection':''}, 
                        'actions':["find", 
                                   "listCollections", 
                                   "createIndex"]}
    db.command('createRole', 'guest', 
               privileges=[guest_privileges], roles=[])

    custom_data = {'registration_date': now}
    db.add_user('guest', password, customData=custom_data, 
                roles=['guest'])

def add_user(name, password):
    '''
    Add a new user with the given name and password
    '''

    now = datetime.datetime.utcnow()

    general_privileges = {'resource':{'db':db_name, 'collection':''}, 
                          'actions':["find", 
                                     "changeOwnPassword", 
                                     "listCollections", 
                                     "createIndex"]}
    user_privileges = {'resource':{'db':db_name, 'collection':name}, 
                       'actions':["find", 
                                  "update", 
                                  "insert", 
                                  "remove"]}
    db.command('createRole', name, 
               privileges=[general_privileges, user_privileges], roles=[])
    custom_data = {'registration_date': now}

    db.add_user(name, password, customData=custom_data, 
                roles=[name], mechanisms=['SCRAM-SHA-1'])

def remove_user(name):
    '''
    Delete a user from the tios database
    '''

    db.command('dropRole', name)
    db.drop_collection(name)
    db.remove_user(name)

def list_users():
    '''
    List the users in the tios database
    '''

    result = db.command('usersInfo', 1)
    return result['users']



import unittest
from tios import namd_utilities
import tempfile
import os
import zlib
import mdtraj as mdt

class TestNAMDUtilitiesMethods(unittest.TestCase):

    def test_config_file_valid_for_imd(self):
        configfile = 'test/examples/ubq_wb.conf'
        self.assertTrue(namd_utilities.config_file_valid_for_imd(configfile))

    def test_config_file_does_not_set_imd(self):
        configfile = 'test/examples/ubq_wb_no_imd.conf'
        self.assertFalse(namd_utilities.config_file_valid_for_imd(configfile))

    def test_config_file_is_NPT(self):
        configfile = 'test/examples/ubq_wb_NPT.conf'
        self.assertFalse(namd_utilities.config_file_valid_for_imd(configfile))

    def test_dt_from_config_file(self):
        configfile = 'test/examples/ubq_wb.conf'
        dt = namd_utilities.dt_from_config_file(configfile)
        self.assertEqual(dt, 0.002)

    def test_trate_from_config_file(self):
        configfile = 'test/examples/ubq_wb.conf'
        trate = namd_utilities.trate_from_config_file(configfile)
        self.assertEqual(trate, 500)

    def test_trajectory_from_config_file(self):
        configfile = 'test/examples/ubq_wb.conf'
        traj = namd_utilities.trajectory_from_config_file(configfile)
        self.assertIsInstance(traj, mdt.Trajectory)
        self.assertEqual(traj.topology.n_atoms, 7051)

    def test_filenames_from_config_file(self):
        configfile = 'test/examples/ubq_wb.conf'
        filenames = namd_utilities.filenames_from_config_file(configfile)
        expected_filenames = ['test/examples/ubq_wb.psf', 
                              'test/examples/ubq_wb.pdb', 
                              'test/examples/ubq_wb_eq.coor', 
                              'test/examples/ubq_wb_eq.xsc', 
                              'test/examples/ubq_wb.conf', 
                              'test/examples/par_all27_prot_lipid.inp']
        #self.assertItemsEqual(filenames, expected_filenames)
        for f in expected_filenames:
            self.assertIn(f, filenames)



"""
Utilities to extract required information from GROMACS files
"""
from __future__ import print_function
import sys
import os
import subprocess
import tempfile
import zlib
import mdtraj as mdt
from tios import environments

def complete_inputs(inputs):
    """
    Complete entries in an inputs dictionary for a Gromacs simulation.

    Used when the invocation has used the "-deffnm" shorthand.

    Args:
        inputs (dict): Dictionary of ``{inputflag : inputvalue}`` pairs.

    Returns:
        dict: Completed dictionary of `inputs`.
    """
    if '-deffnm' in inputs:
        deffnm = inputs['-deffnm']
        if not '-s' in inputs:
            inputs['-s'] = deffnm + '.tpr'
        if not '-o' in inputs:
            inputs['-o'] = deffnm + '.trr'
        if not '-x' in inputs:
            inputs['-x'] = deffnm + '.xtc'
        if not '-cpo' in inputs:
            inputs['-cpo'] = deffnm + '.cpt'
        if not '-e' in inputs:
            inputs['-e'] = deffnm + '.edr'
        if not '-g' in inputs:
            inputs['-g'] = deffnm + '.log'
        del inputs['-deffnm']
    return inputs

def complete_inputs_for_imd(inputs):
    """
    Checks and amends the inputs dictionary for a valid gromacs IMD run.

    Args:
        inputs (dict): Dictionary of ``{inputflag : inputvalue}`` pairs.

    Returns:
        dict: Completed dictionary of `inputs`.
    """
    new_inputs = inputs.copy()
    env = environments.load_mongo_env()
    if not '-imdwait' in new_inputs:
        new_inputs['-imdwait'] = None
    if not '-imdterm' in new_inputs:
        new_inputs['-imdterm'] = None
    if not '-imdport' in new_inputs:
        new_inputs['-imdport'] = str(env['IMD_PORT'])
    return new_inputs

def apply_checkpoint(inputs, checkpoint):
    """
    Unpack checkpoint data and update the inputs dictionary accordingly.

    Args:
        inputs (dict): Dictionary of ``{inputflag : inputvalue}`` pairs.
        checkpoint (str): ``zlib`` compressed checkpoint file contents.

    Returns:
        dict: Completed dictionary of `inputs`
    """
    new_inputs = inputs.copy()
    if '-cpo' in inputs:
        checkpointfile = inputs['-cpo']
    elif '-deffnm' in inputs:
        checkpointfile = inputs['-deffnm'] + '.cpt'
    elif '-s' in inputs:
        checkpointfile = os.path.splitext(inputs['-s'])[0] +'.cpt'
    else:
        checkpointfile = 'system.cpt'

    new_inputs['-cpi'] = checkpointfile
    with open(checkpointfile, 'wb') as f:
        f.write(zlib.decompress(checkpoint))

    new_inputs['-noappend'] = None
    return new_inputs
       
def tpr_file_valid_for_imd(tprfile):
    """
    Checks that the tpr file contains all required parameters for an IMD run.

    Args:
        tprfile (str): Name of GROMACS tpr file.

    Returns:
        bool: True if the file is valid, else False.

    """
    if not os.path.exists(tprfile):
        print('Error: cannot find {}'.format(tprfile))
        return False
    try:
        result = subprocess.check_output(["gmx", "dump", "-s",
                                          tprfile],
                                          stderr=subprocess.STDOUT)
    except:
        message = sys.exc_info()[1]
        print('Error:', message)
        if isinstance(message, tuple):
            errno = sys.exc_info()[1][0]
            if errno == 2:
                print('Is the gmx command in your path?')
        return False

    pcoupl = None
    for line in result.split(b'\n'):
        w = line.split()
        if len(w) > 0:
            if w[0] == b'pcoupl':
                pcoupl = w[2]
    if pcoupl != b'No' and pcoupl != b'no':
        print('Error: Tios does not support constant pressure simulations')
        print('Please set pcoupl = no in your .mdp file and re-run grompp')
        return False

    if not b'IMD-atoms' in result:
        print("Error: IMD has not been enabled for this simulation.")
        print("Please add \"imd-group = System\" to your .mdp file and")
        print("re-run grompp.")
        return False

    return True

def dt_from_tpr_file(tprfile):
    """
    Extracts the MD timestep from the tpr file.

    Args:
        tprfile (str): Name of the Gromacs tpr file.

    Returns:
        float: The basic MD timestep unit (in picoseconds).

    Raises:
        RuntimeError: if the value of `dt` cannot be determined.

    """
    dt = None
    try:
        result = subprocess.check_output(["gmx", "dump", "-s",
                                          tprfile],
                                          stderr=subprocess.STDOUT)
    except:
        print('Error:', sys.exc_info()[1])
        errno = sys.exc_info()[1][0]
        if errno == 2:
            print('Is the gmx command in your path?')
        exit(1)

    for line in result.split(b'\n'):
        w = line.split()
        if len(w) > 0:
            if w[0] == b'dt':
                dt = float(w[2])
    if dt is None:
        raise RuntimeError('Error: cannot find MD timestep info in .tpr file')
    return dt

def trajectory_from_tpr_file(tprfile):
    """
    Create an MDTraj Trajectory from the tprfile.

    Args:
        tprfile (str): Name of the Gromacs tpr file.

    Returns:
        MDTraj.Trajectory: Trajectory object. 

    """
    cwd = os.getcwd()
    try:
        topfile = os.path.basename(tempfile.NamedTemporaryFile(dir=cwd, suffix='.pdb', delete=False).name)
        result2 = None
        result2 = subprocess.check_output(["gmx", "editconf", "-f",
                                          tprfile, "-o", topfile,
                                          "-conect"],
                                          stderr=subprocess.STDOUT)
    except:
        print('Error:', sys.exc_info())
        print(result2)
        exit(1)
    trajectory = mdt.load(topfile)
    os.remove(topfile)
    return trajectory

def step_from_checkpoint_file(cptfile):
    """
    Extracts the current time step from a checkpoint file.

    Args:
        cptfile (str): Name of the checkpoint file.

    Returns:
        int: The final timestep (in units of `dt`).

    Raises:
        RuntimeError: if `cptfile` cannot be found, or the current
            timestep cannot be identified.

    """
    step = None
    if not os.path.exists(cptfile):
        raise RuntimeError('Error: cannot find {}'.format(cptfile))
    try:
        result = subprocess.check_output(["gmx", "dump", "-cp",
                                          cptfile],
                                          stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print(e.output, e.returncode)
        exit(1)
    except OSError as e:
        if e.errno == 2:
            print('Is the gmx command in your path?')
        else:
            print(e)
        exit(1)
    except:
        message = sys.exc_info()[1]
        print('Error:', message)
        exit(1)

    for line in result.split(b'\n'):
        w = line.split()
        if len(w) > 0:
            if w[0] == b'step':
                step = int(w[2])
    if step is None:
        raise RuntimeError('Error: cannot find current step info in .cpt file')
    return step

def gromacs_check_version():
    '''
    Return the version of Gromacs available on this resource.

    Returns:
        The Gromacs version as a string, or None if it cannot be determined.

    '''
    gmx = gromacs_find_gmx()
    if gmx is None:
        return None
    try:
        output = subprocess.check_output([gmx, '-version'], stderr=subprocess.STDOUT, universal_newlines=True)
    except:
        return None
    version = 'unknown'
    for line in output.split():
        words = line.split()
        if 'version:' in words:
            version = words[-1]
            break

    return version

def gromacs_find_gmx():
    '''
    Find out if the current environment has 'gmx' or 'gmx_d' commands available.

    Returns:
        "gmx", "gmx_d", or None.

    '''
    gmx = None
    try:
        output = subprocess.check_output(['which', 'gmx'], stderr=subprocess.STDOUT)
        gmx = 'gmx'
    except:
        try:
            output = subprocess.check_output(['which', 'gmx_d'], stderr=subprocess.STDOUT)
            gmx = 'gmx_d'
        except:
            pass
    return gmx
        



import os

try:
    import popylar
    popylar.track_event('UA-102907451-1', 'import', 'import_tios')
except ImportError:
    pass

TIOS_CONFIGDIR = os.getenv('TIOS_CONFIGDIR', 
                           os.path.join(os.getenv('HOME'), '.tios'))
TIOS_CONFIGFILE = os.path.join(TIOS_CONFIGDIR, 'tios.cfg')
if os.path.exists(TIOS_CONFIGFILE):
    permissions = os.stat(TIOS_CONFIGFILE).st_mode & 0o777
    if permissions & 0o077 > 0:
        message = ('Error - your tios configuration file {} '+
                   'has permissions {} - it must be readable ' +
                   'only by you.')
        raise RuntimeError(message.format(TIOS_CONFIGFILE, 
                                          oct(permissions & 0o777)))




import unittest
from tios import transmitter, communication
import shutil
import tempfile
import os
import mdtraj as mdt

class TestTiosTransmitterMethods(unittest.TestCase):

    def setUp(self):
        #self.protocol = 'Dummy'
        self.protocol = 'Mongo'
        #self.testdir = tempfile.mkdtemp()
        self.testdir = 'scratchdir'
        if os.path.exists(self.testdir):
            shutil.rmtree(self.testdir)
        os.mkdir(self.testdir)
        shutil.copy('test/examples/bpti.tpr', self.testdir)
        shutil.copy('test/examples/bpti.cpt', self.testdir)

    def tearDown(self):
        shutil.rmtree(self.testdir)

    def test_initialize_new_tios_transmitter_from_command(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol = self.protocol)
        self.assertIsInstance(tj, transmitter.TiosTransmitter)

    def test_initialize_new_tios_transmitter_from_database(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        id = tj.id

        tj2 = transmitter.from_entry(id, protocol=self.protocol, targetdir=self.testdir)
        self.assertIsInstance(tj2, transmitter.TiosTransmitter)

    def test_tios_job_from_command_datastore_integrity(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        self.assertEqual(tj._te.title, 'test')
        self.assertEqual(tj._te.timepoint, 0.0)
        self.assertEqual(tj._te.md_code, 'GROMACS')
        self.assertEqual(tj._te.status, 'Ready')
        self.assertEqual(tj._te.trate, 1.0)
        self.assertEqual(tj._te.frame_rate, 0.0)
        self.assertIsInstance(tj._te.topology, mdt.Topology)
        self.assertEqual(len(tj._te.selection), 58)
        self.assertEqual(len(tj._te.xyzsel), 58)
        self.assertEqual(len(tj._te.xyzsel[0]), 3)

    def test_tios_job_from_command_datastore_integrity2(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        id = tj.id

        tj2 = transmitter.from_entry(id, protocol=self.protocol, targetdir=self.testdir)

        self.assertEqual(tj2._te.title, 'test')
        self.assertEqual(tj2._te.timepoint, 0.0)
        self.assertEqual(tj2._te.md_code, 'GROMACS')
        self.assertEqual(tj2._te.status, 'Ready')
        self.assertEqual(tj2._te.trate, 1.0)
        self.assertEqual(tj2._te.frame_rate, 0.0)
        self.assertIsInstance(tj2._te.topology, mdt.Topology)
        self.assertEqual(len(tj2._te.selection), 58)
        self.assertEqual(tj2._te.xyzsel.shape, (58, 3))

    def test_tios_transmitter_start_step_and_stop(self):
    
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        tj.start()
        self.assertEqual(tj._te.status, 'Running')
        self.assertFalse(tj._imd_sim.new_checkpoint)
        tj.step()
        self.assertEqual(tj._te.timepoint, 0.002)
        tj.step()
        self.assertEqual(tj._te.timepoint, 1.002)
        tj.stop()
        self.assertEqual(tj._te.status, 'Stopped')
        
    def test_tios_transmitter_restart_step_and_stop(self):
    
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti -cpi {}/bpti.cpt -noappend'.format(self.testdir, self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        self.assertEqual(tj._te.status, 'Ready')
        self.assertEqual(tj._te.timepoint, 1.04)
        self.assertTrue('-cpi' in tj._te.filepack)
        tj.start()
        self.assertEqual(tj._te.status, 'Running')
        tj.step()
        #self.assertEqual(tj._te.timepoint, 1.202)
        tj.step()
        self.assertAlmostEqual(tj._te.timepoint, 2.002)
        tj.stop()
        self.assertEqual(tj._te.status, 'Stopped')

    def test_tios_transmitter_start_stop_and_restart(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        tj.start()
        tj.step()
        tj.step()
        tj.stop()
        self.assertAlmostEqual(tj._te.timepoint, 1.002)
        id = tj.id

        tj2 = transmitter.from_entry(id, protocol=self.protocol, targetdir=self.testdir)
        #self.assertAlmostEqual(tj2._te.timepoint, 1.012)
        tj2.start()
        self.assertEqual(tj2._te.status, 'Running')
        tj2.stop()
        self.assertEqual(tj2._te.status, 'Stopped')

    def test_tios_transmitter_stop_by_message(self):
        tj = transmitter.from_command('gmx mdrun -deffnm {}/bpti'.format(self.testdir), 
                                   title='test',
                                   selection='name CA', 
                                   protocol=self.protocol)
        tj.start()
        tj.step()
        tj.step()
        self.assertEqual(tj._te.status, 'Running')
        tj._te.message = {'header' : 'STOP'}
        tj._te.sync()
        tj.step()
        self.assertEqual(tj._te.status, 'Stopped')




import unittest
import shutil
import tempfile
from tios import imd_sim
import glob
import os
import time
import numpy as np
import mdtraj as mdt
import sys
import warnings

class NAMDWorkflowTest(unittest.TestCase):

    def setUp(self):
        #self.testdir = tempfile.mkdtemp()
        self.testdir = 'scratchdir'
        if os.path.exists(self.testdir):
            shutil.rmtree(self.testdir)
        os.mkdir(self.testdir)
        shutil.copy('test/examples/ubq_wb.conf', self.testdir)
        shutil.copy('test/examples/ubq_wb_eq.coor', self.testdir)
        shutil.copy('test/examples/ubq_wb_eq.xsc', self.testdir)
        shutil.copy('test/examples/ubq_wb.psf', self.testdir)
        shutil.copy('test/examples/ubq_wb.pdb', self.testdir)
        shutil.copy('test/examples/par_all27_prot_lipid.inp', self.testdir)
        if sys.version_info[0] > 2:
            warnings.simplefilter("ignore", ResourceWarning)

    def tearDown(self):
        #shutil.rmtree(self.testdir)
        pass

    def test_simulation_startup(self):
        command_line = 'namd2 +p4 {}/ubq_wb.conf'.format(self.testdir)
        # Attempt to start the job:
        imd_job = imd_sim.from_command(command_line)
        # Have we created an instance of an imd_sim.imd_job?
        self.assertIsInstance(imd_job, imd_sim.IMDJob)
        # Is the status 'Ready'?
        self.assertEqual(imd_job.status, 'Ready')
        self.assertEqual(imd_job.timepoint, 0.0)
        self.assertIsInstance(imd_job.trajectory, mdt.Trajectory)
        self.assertIsInstance(imd_job.trajectory.xyz, np.ndarray)

    def test_simulation_start(self):
        command_line = 'namd2 +p4 {}/ubq_wb.conf'.format(self.testdir)
        imd_job = imd_sim.from_command(command_line)
        imd_job.start()
        # Check we are at timepoint 0.0:
        self.assertEqual(imd_job.timepoint, 0.0)
        # Move the simulation forward by one step. First time
        # it should only be advanced by dt:
        imd_job.step()
        self.assertEqual(imd_job.timepoint, imd_job._trate)
        # Move the simulation forward again. This time it should be trate
        # further forward:
        imd_job.step()
        self.assertEqual(imd_job.timepoint, 2 * imd_job._trate)
        # Stop the job and confirm that it appears to be so:
        imd_job.stop()
        time.sleep(3)
        self.assertFalse(imd_job.is_running())
        # A checkpoint file should have been written - has it?
        self.assertTrue(imd_job._has_checkpoint)
        self.assertTrue(imd_job.new_checkpoint)
        # Retrieve the checkpoint and test new_checkpoint again:
        cpt = imd_job.get_checkpoint()
        self.assertFalse(imd_job.new_checkpoint)
      



import os
import json
import pymongo
from pymongo import MongoClient
import tios
"""
Module with functions to initialise and edit the Tios environment

Currently only a Mongo-based environment is available.
"""
def load_mongo_env(source=None, authenticate=True, administrator=False):
    '''
    Get the tios environment from the configuration file, if it exists.
    '''
    if source is None:
        if os.path.exists(tios.TIOS_CONFIGFILE):
            with open(tios.TIOS_CONFIGFILE, 'r') as f:
                env = json.load(f)
        else:
            env = {
                    'DB_URL': 'tirith.pharm.nottingham.ac.uk',
                    'DB_NAME': 'tios',
                    'DB_COLLECTION': 'guest',
                    'DB_PORT': 27017,
                    'DB_USER': 'guest',
                    'DB_PWD': 'guest123',
                    'IMD_PORT': '40237',
                  }
        save_mongo_env(env)
    else:
        env = source

    if not authenticate:
        return env

    db = MongoClient(env['DB_URL'], int(env['DB_PORT']))[env['DB_NAME']]
    if administrator:
        if 'DB_ADMINISTRATOR' in env and 'DB_ADMIN_PWD' in env:
            db.authenticate(name=env['DB_ADMINISTRATOR'], 
                            password=env['DB_ADMIN_PWD'], source='admin')
        else:
            raise RuntimeError('You do not have administrator rights')
    else:
        db.authenticate(name=env['DB_USER'], password=env['DB_PWD'])
    env['COLLECTION'] = db[env['DB_COLLECTION']]
    return env

def save_mongo_env(env):
    '''
    Save the tios environment to the configuration file.
    '''
    coll = None
    if 'COLLECTION' in env:
        coll = env.pop('COLLECTION')
    if not os.path.exists(tios.TIOS_CONFIGDIR):
        os.mkdir(tios.TIOS_CONFIGDIR)
    with open(tios.TIOS_CONFIGFILE, 'w') as f:
        json.dump(env, f, indent=4)
    os.chmod(tios.TIOS_CONFIGFILE, 0o600)
    if coll is not None:
        env['COLLECTION'] = coll




#!/usr/bin/env python
'''
Create the web page for a simulation.
'''
import sys
from tios import environments, communication, admin, utilities
import tios.graphics as tg
import pickle
import mdtraj as mdt
import tempfile
import os
import os.path as op
import pymongo
import datetime

basedir = '/users/charlie/public_html'
tmpdir = '/users/charlie/public_html/tmp'
monwid = 4.0
monht = 0.2


def id_to_html(id, cname):
    '''
    Create the html for a web page for tios simulation with the given id
    '''

    te = communication.TiosAgent(id)
    monitors = te.monitors
    status = te.status
    offset = datetime.datetime.now() - datetime.datetime.utcnow()
    last_update = (te.last_update + offset).strftime("%d/%m/%y %H:%M:%S")
    title = te.title
    frame_rate = te.frame_rate
    interval = te.trate
    xyz = te.xyzsel
    md_code = te.md_code

    imgfile = tempfile.NamedTemporaryFile(suffix='.png', dir=tmpdir, delete=False).name

    top = te.topology
    n_atoms = top.n_atoms
    sel = te.selection
    top = top.subset(sel)
    t = mdt.Trajectory(xyz, top)
    t.unitcell_vectors = te.box
    t.topology.create_standard_bonds()
    t.make_molecules_whole()
    sel2 = t.topology.select('not type H')
    t.topology = t.topology.subset(sel2)
    t.xyz = t.xyz[:, sel2]

    for key in monitors:
        if not isinstance(monitors[key], str):
            monitors[key] = monitors[key][-1]

    imgwid = max(monwid, monht * len(monitors))
    tg.plot_molecule(imgfile, t, size=(imgwid,imgwid))
    os.chmod(imgfile, 0644)

    result = ''
    result += '<head></head>\n'
    result +='<table><tr><td rowspan=2><img src=\"../tios3.gif\"></td><td><h2>TIOS: '
    result +='Status page for simulation {} ({})</h2></td></tr>\n'.format(id, status)
    result +='<tr><td><h3>{}</h3></td></tr></table>\n'.format(title)
    result += "<table><tr><td>Last update:</td><td>{}</td></tr>".format(last_update)
    result += "<tr><td>Number of atoms:</td><td>{}</td></tr>".format(n_atoms)
    result += "<tr><td>MD code:</td><td>{}</td></tr>".format(md_code)
    result += "<tr><td>Current frame rate (/min):</td><td>{:4.1f}</td></tr>".format(frame_rate)
    result += "<tr><td>Sampling interval (ps):</td><td>{:4.0f}</td></tr></table>".format(interval)
    result += '<table><tr><td rowspan={}>'.format(len(monitors))
    result += '<img src=\"../tmp/{}\">'.format(op.basename(imgfile))
    result += '</td>\n'
    result += '<td>Timepoint</td><td>{:10.5g}</td></tr>\n'.format(float(monitors['Timepoint']))
    del monitors['Timepoint']
    for key in monitors:
        result += '<tr><td>{}</td><td>{:10.5g}</td></tr>\n'.format(key, float(monitors[key]))
    result += '</table>\n'
    return result

def register(user, email):
    '''
    Generates the web page that informs a new user of their registration details
    '''

    tios_env = environments.load_mongo_env()
    result = ''
    result += '<head></head>\n'
    result += '<center><img src=\"../tios2.gif\"></center>'
    result += '<center><h1>TIOS: The Internet of Simulations</h1></center><p>'
    
    if user is None or email is None:
        result += '<b>Error: you must supply both a user name and an email address</b><br>\n'
        return result
    try:
        password = utilities.id_generator(size=8)
        admin.add_user(user, password)
    except pymongo.errors.DuplicateKeyError:
        result += '<b>Error - this user name has already been taken.<br>\n' 
        result += 'Please try another one</b><br>\n'
        return result
    result += '<h2>Your tios registration is complete</h2><p>\n'
    result += 'Now run \'tios-config\' on each resource you use with the '
    result += 'following settings:<p>'
    result += '<table><tr><td><b>Database URL:</b></td><td>' + tios_env['DB_URL'] + '</td></tr>'
    result += '<tr><td><b>Database Port:</b></td><td>{}</td></tr>'.format( tios_env['DB_PORT'])
    result += '<tr><td><b>Database name:</b></td><td>' + tios_env['DB_NAME'] + '</td></tr>'
    result += '<tr><td><b>Username:</b></td><td>' + user + '</td></tr>'
    result += '<tr><td><b>Password:</b></td><td>' + password + '</td></tr></table><p>'
    result += ' (you can change your password using \'tios-passwd\') <br>'
    return result



# Python module suppprting the Interactive Molecular Dynamics protocol
import socket
import struct

IMD_DISCONNECT = 0
IMD_ENERGIES = 1
IMD_FCOORDS = 2
IMD_GO = 3
IMD_HANDSHAKE = 4
IMD_KILL = 5
IMD_MDCOMM = 6
IMD_PAUSE = 7
IMD_TRATE = 8
IMD_IOERROR = 9
IMD_ETERMS = ['tstep', 'T', 'Etot', 'Epot', 
              'Evdw', 'Eelec', 'Ebond', 
              'Eangle', 'Edihe', 'Eimpr']

HEADERSIZE = 8
IMDVERSION = 2


def imd_connect(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
    except socket.error:
        #print('Error: host: {}, port: {}'.format(host, port))
        raise
    return s

def imd_readn(s, n):
    buff = b''
    bytes_recd = 0
    while bytes_recd < n:
        chunk = s.recv(min(n - bytes_recd, 2048))
        if len(chunk) == 0:
            raise RuntimeError("socket connection broken")
        buff += chunk
        bytes_recd += len(chunk)
    return buff

def imd_send_header(s, typ, length):
    msg = struct.pack('II', socket.htonl(typ), socket.htonl(length))
    s.sendall(msg)

def imd_disconnect(s):
    imd_send_header(s, IMD_DISCONNECT, 0)
    
def imd_pause(s):
    imd_send_header(s, IMD_PAUSE, 0)

def imd_kill(s):
    try:
        imd_send_header(s, IMD_KILL, 0)
    except socket.error:
        pass

def imd_go(s):
    imd_send_header(s, IMD_GO, 0)

def imd_handshake(s):
    msg = struct.pack('II', socket.htonl(IMD_HANDSHAKE), IMDVERSION)
    s.sendall(msg)

def imd_trate(s, rate):
    imd_send_header(s, IMD_TRATE, int(rate))

def imd_send_mdcomm(s, n_forces, indices, forces):
    imd_send_header(IMD_MDCOMM, n_forces)
    msg = struct.pack('I' * n_forces + 'f' * n_forces * 3, *(indices, forces))
    s.sendall(msg)

def imd_send_energies(s, energies):
    imd_send_header(IMD_ENERGIES, 1)
    msg = struct.pack('Ifffffffff', *[energies[eterm] for eterm in IMD_ETERMS])
    s.sendall(msg)

def imd_send_fcoords(s, n_atoms, coords):
    imd_send_header(IMD_FCOORDS, n_atoms)
    msg = struct.pack('f' * 3 * n_atoms, *coords)
    s.sendall(msg)

def imd_recv_header(s):
    buff = imd_readn(s, HEADERSIZE)
    typ, length = struct.unpack('II', buff)
    typ = socket.ntohl(typ)
    if typ != IMD_HANDSHAKE:
        length = socket.ntohl(length)
    return typ, length

def imd_recv_handshake(s):
    typ, length = imd_recv_header(s)
    if typ != IMD_HANDSHAKE:
        return -1
    if length == IMDVERSION:
        imd_go(s)
        return 0
    length = struct.unpack("<I", struct.pack(">I", length))[0]
    if length == IMDVERSION:
        imd_go(s)
        return 1
    return -1

def imd_recv_mdcomm(s, n_forces):
    imd_readn(s, buff, n_forces * 4 * 4)
    indfor = struct.unpack('I' * n_forces + 'fff' * n_forces, buff)
    indices = indfor[:n_forces]
    forces = indfor[n_forces:]
    return indices, forces

def imd_recv_energies(s):
    buff = imd_readn(s, 40)
    eners = struct.unpack('Ifffffffff', buff)
    energies = {}
    for i in range(len(IMD_ETERMS)): 
        energies[IMD_ETERMS[i]] = eners[i]
    return energies

def imd_recv_fcoords(s, n_atoms):
    buff = imd_readn(s, 12 * n_atoms)
    coords = struct.unpack('fff' * n_atoms, buff)
    return coords




import unittest
from tios import utilities

class TestCheckGromacs2018Methods(unittest.TestCase):

    def test_check_gromacs_installed_version(self):
        result = utilities.installed_version('GROMACS')
        self.assertNotEqual(result, None)
